from scipy import sparse as sp
import pandas as pd
import numpy as np
import bw2data as bd
from bw_temporalis import TemporalDistribution
from .remapping import TimeMappingDict
from bw2calc import LCA
from datetime import datetime
from .utils import convert_date_string_to_datetime


class DynamicBiosphereBuilder:
    """
    This class is used to build a dynamic biosphere matrix, which in contrast to the normal biosphere matrix has rows for each biosphere flow at their time of emission. 
    Thus, the dimensions are (bio_flows at a specific timestep) x (processes).
    """

    def __init__(
        self,
        lca_obj: LCA,
        activity_time_mapping_dict: dict,
        biosphere_time_mapping_dict: dict,
        demand_timing_dict: dict,
        node_id_collection_dict: dict,
        temporal_grouping: str,
        database_date_dict: dict,
        database_date_dict_static_only: dict,
    ) -> None:
        """ 
        Initializes the DynamicBiosphereBuilder object.

        Parameters
        ----------
        lca_obj : LCA objecct
            instance of the bw2calc LCA class, e.g. TimexLCA.lca
        activity_time_mapping_dict : dict
            A dictionary mapping activity to their respective timing in the format (('database', 'code'), datetime_as_integer): time_mapping_id)
        biosphere_time_mapping_dict : dict
            A dictionary mapping biosphere flows to their respective timing in the format (('database', 'code'), datetime_as_integer): time_mapping_id), at this point still empty.
        demand_timing_dict : dict
            A dictionary mapping of the demand to demand time
        node_id_collection_dict : dict
            A dictionary containing lists of node ids for different node subsets
        temporal_grouping : str
            A string indicating the temporal grouping of the processes, e.g. 'year', 'month', 'day', 'hour'
        database_date_dict : dict
            A dictionary mapping database names to their respective date
        database_date_dict_static_only : dict
            A dictionary mapping database names to their respective date, but only containing static databases, which are the background databases.
 
        Returns
        -------
        None

        """

        self._time_res_dict = {
            "year": "datetime64[Y]",
            "month": "datetime64[M]",
            "day": "datetime64[D]",
            "hour": "datetime64[h]",
        }

        self.lca_obj = lca_obj
        self.technosphere_matrix = lca_obj.technosphere_matrix
        self.activity_dict = lca_obj.dicts.activity
        self.activity_time_mapping_dict = activity_time_mapping_dict
        self.biosphere_time_mapping_dict = biosphere_time_mapping_dict
        self.demand_timing_dict = demand_timing_dict
        self.node_id_collection_dict = node_id_collection_dict
        self.time_res = self._time_res_dict[temporal_grouping]
        self.temporal_grouping = temporal_grouping
        self.database_date_dict = database_date_dict
        self.database_date_dict_static_only = database_date_dict_static_only
        self.dynamic_supply_array = lca_obj.supply_array
        self.rows = []
        self.cols = []
        self.values = []

    def build_dynamic_biosphere_matrix(self):
        """
        This function creates a separate biosphere matrix, with the dimenions (bio_flows at a specific timestep) x (processes).
        Every temporally resolved biosphere flow has its own row in the matrix, making it highly sparse.
        The timing of the emitting process and potential additional temporal information of the bioshpere flow (e.g. delay of emission compared to timing of process) are considered.
        
        Parameters
        ----------
        None

        Returns
        -------
        dynamic_biomatrix : scipy.sparse.csr_matrix
            A sparse matrix with the dimensions (bio_flows at a specific timestep) x (processes), where every row represents a biosphere flow at a specific time.
        """

        for id in self.node_id_collection_dict["temporalized_processes"]:
            process_col_index = self.activity_dict[id]  # get the matrix column index

            (
                (original_db, original_code),
                time,
            ) = self.activity_time_mapping_dict.reversed()[  # time is here an integer, with various length depending on temporal grouping, e.g. [Y] -> 2024, [M] - > 202401
                id
            ]

            time_in_datetime = convert_date_string_to_datetime(
                self.temporal_grouping, str(time)
            )  # now time is a datetime

            td_producer = TemporalDistribution(
                date=np.array([time_in_datetime], dtype=self.time_res),
                amount=np.array([1]),
            ).date
            date = td_producer[0]

            act = bd.get_node(database=original_db, code=original_code)

            for exc in act.biosphere():
                if exc.get("temporal_distribution"):
                    td_dates = exc["temporal_distribution"].date  # time_delta
                    td_values = exc["temporal_distribution"].amount
                    dates = (
                        td_producer + td_dates
                    )  # we can add a datetime of length 1 to a timedelta of length N without problems
                    values = exc["amount"] * td_values

                else:  # exchange has no TD
                    dates = td_producer  # datetime array, same time as producer
                    values = [exc["amount"]]

                # Add entries to dynamic bio matrix
                for date, amount in zip(dates, values):

                    # first create a row index for the tuple((db, bioflow), date))
                    time_mapped_matrix_id = self.biosphere_time_mapping_dict.add(
                        (exc.input, date)
                    )

                    # populate lists with which sparse matrix is constructed
                    self.add_matrix_entry_for_biosphere_flows(
                        row=time_mapped_matrix_id,
                        col=process_col_index,
                        amount=amount,
                    )

        for id in self.node_id_collection_dict["temporal_markets"]:
            process_col_index = self.activity_dict[id]  # get the matrix column index
            technosphere_column = (
                self.technosphere_matrix[:, process_col_index].toarray().flatten()
            )  # 1-d np.array
            demand = dict()
            for idx, amount in enumerate(technosphere_column):
                if idx == self.activity_dict[id]:  # Skip production exchange
                    continue
                if amount == 0:
                    continue

                node_id = self.activity_dict.reversed[idx]

                if (
                    node_id in self.node_id_collection_dict["foreground_node_ids"]
                ):  # We only aggregate background process bioflows
                    continue

                # demand[bd.get_node(id=node_id)] = -amount
                demand[node_id] = -amount

            self.lca_obj.redo_lci(demand)
            aggregated_inventory = self.lca_obj.inventory.sum(
                axis=1
            )  # aggregated biosphere flows of background supply chain emissions. Rows are bioflows.

            for idx, amount in enumerate(aggregated_inventory.flatten().tolist()[0]):
                bioflow = bd.get_activity(self.lca_obj.dicts.biosphere.reversed[idx])
                ((_, _), time) = self.activity_time_mapping_dict.reversed()[id]

                time_in_datetime = convert_date_string_to_datetime(
                    self.temporal_grouping, str(time)
                )  # now time is a datetime

                td_producer = TemporalDistribution(
                    date=np.array([str(time_in_datetime)], dtype=self.time_res),
                    amount=np.array([1]),
                ).date  # TODO: Simplify
                date = td_producer[0]

                time_mapped_matrix_id = self.biosphere_time_mapping_dict.add(
                    (bioflow, date)
                )

                self.add_matrix_entry_for_biosphere_flows(
                    row=time_mapped_matrix_id, col=process_col_index, amount=amount
                )

        # now build the dynamic biosphere matrix
        shape = (max(self.rows) + 1, len(self.activity_time_mapping_dict))
        dynamic_biomatrix = sp.coo_matrix((self.values, (self.rows, self.cols)), shape)
        self.dynamic_biomatrix = dynamic_biomatrix.tocsr()

        return self.dynamic_biomatrix

    def add_matrix_entry_for_biosphere_flows(self, row, col, amount):
        """
        Adds an entry to the lists of row, col and values, which are then used to construct the dynamic biosphere matrix.

        Parameters
        ----------
        row : int
            A row index of a new element to the dynamic biosphere matrix
        col: int
            A column index of a new element to the dynamic biosphere matrix
        amount: float
            The amount of the new element to the dynamic biosphere matrix

        Returns
        -------
        None, but the lists of row, col and values are updated

        """
        self.rows.append(row)
        self.cols.append(col)
        self.values.append(amount)



"""
unit tests for testing if consecutive distributions are fine:
 test 1: test_two_consecutive_TD_in_the_past_direct_match_db
 test 2: test_two_consecutive_TD_in_the_past_interpol_db_no_overlap

"""

import math
import unittest
from bw_temporalis import easy_timedelta_distribution, TemporalDistribution
from edge_extractor import EdgeExtracter
from medusa_tools import *
import bw2data as bd
import bw2calc as bc
import numpy as np


class TestTemporalDistributions(unittest.TestCase):
    
    def test_two_consecutive_TD_in_the_past_direct_match_db(self):
        ''' 
        Test if two consecutive temporal distributions in the past are correctly aggregated
        TD has been selected to match time of the background databases directly, so no interpolation is needed
        Only unitary processes implemented
        
        '''
        
        def prepare_medusa_dbs():
            bd.projects.set_current("__test_medusa__")
            
            bd.Database('temporalis-bio').write({
                ('temporalis-bio', "CO2"): {
                    "type": "emission",
                    "name": "carbon dioxide",
                    "temporalis code": "co2",
                },
                ('temporalis-bio', "CH4"): {
                    "type": "emission",
                    "name": "methane",
                    "temporalis code": "ch4",
                },
            })

            bd.Database('background_2022').write({
                ('background_2022', 'C'): {
                'name': 'process C',
                "location": "somewhere",
                'reference product': 'C',
                'exchanges': [
                        {
                            'amount': 1,
                            'type': 'production',
                            'input': ('background_2022', 'C'),
                        },
                        {
                            'amount': 10,
                            'type': 'biosphere',
                            'input': ('temporalis-bio', 'CO2'),
                        },  ]},
            },

                    )

            bd.Database('background_2020').write({
                ('background_2020', 'C'): {
                'name': 'process C',
                "location": "somewhere",
                'reference product': 'C',
                'exchanges': [
                        {
                            'amount': 1,
                            'type': 'production',
                            'input': ('background_2020', 'C'),
                        },
                        {
                            'amount': 30,
                            'type': 'biosphere',
                            'input': ('temporalis-bio', 'CO2'),
                        },  ]},
            },

                    )

            bd.Database('foreground').write({
                ('foreground', 'A'): {
                    'name': 'process A',
                    "location": "somewhere",
                    'reference product': 'A',
                    'exchanges': [
                        {
                            'amount': 1,
                            'type': 'production',
                            'input': ('foreground', 'A'),
                        },
                        {
                            'amount': 1,
                            'type': 'technosphere',
                            'input': ('foreground', 'B'),
                            'temporal_distribution': TemporalDistribution(
                                np.array([-2, 0], dtype='timedelta64[Y]'),
                                np.array([0.5, 0.5])),  
                        },
                     
                       
                    ]
                },
                ('foreground', 'B'):
                {
                    "name": "process B",
                    "location": "somewhere",
                    'reference product': 'B',
                    "exchanges": [
                        {
                            'amount': 1,
                            'type': 'technosphere',
                            'input': ('background_2022', 'C'),
                            'temporal_distribution': TemporalDistribution(
                                np.array([-2, -0], dtype='timedelta64[Y]'), 
                                np.array([0.5, 0.5])), 
                        },
                        {
                            'amount': 1,
                            'type': 'production',
                            'input': ('foreground', 'B'), 
                        }
                    ]

                },
                

    })
            
            
            bd.Method(("GWP", "example")).write([
            (("temporalis-bio", "CO2"), 1),
            ])    
            return 
        
        prepare_medusa_dbs()
        bd.projects.set_current("__test_medusa__") # monthly grouping
        bd.databases
        
        demand = {('foreground', 'A'): 1}
        method = ("GWP", "example")
        
        # calculate static LCA'
        slca = bc.LCA(demand, method)
        slca.lci()
        slca.lcia()     
        
        # calculate Medusa LCA
        SKIPPABLE = [node.id for node in bd.Database('background_2022')] + [
                    node.id for node in bd.Database('background_2020')
        ]

        def filter_function(database_id: int) -> bool:
            return database_id in SKIPPABLE
        
        eelca = EdgeExtracter(slca, edge_filter_function=filter_function)
        
        timeline = eelca.build_edge_timeline()
        
        database_date_dict = {
                datetime.strptime("2020-01", "%Y-%m"): 'background_2020',
                datetime.strptime("2022-01", "%Y-%m"): 'background_2022',        
            }

        timeline_df = create_grouped_edge_dataframe(timeline, database_date_dict, temporal_grouping= 'month', interpolation_type="linear") # monthly grouping

        demand_timing_dict = create_demand_timing_dict(timeline_df, demand)

        dp = create_datapackage_from_edge_timeline(timeline_df, database_date_dict, demand_timing_dict)
        
        fu, data_objs, remapping = prepare_medusa_lca_inputs(demand=demand, demand_timing_dict=demand_timing_dict, method=method) 
        
        lca = bc.LCA(fu, data_objs = data_objs + [dp], remapping_dicts=remapping)
        lca.lci()
        lca.lcia()
        
        #calculate expected score:
        expected_score = 0.25 * 30 + 0.5 * 30 + 0.25 * 10   
        
        print('\nTest test two consecutive TD in the past, direct match with db (weights are only 1, no shares):')
        print(timeline_df)
        print('Static LCA Score:', slca.score)
        print('MEDUSA LCA Score:', lca.score)
        print('Expected MEDUSA LCA Score:', expected_score)
        
        
        # Check if the results are equal using math.isclose
        self.assertTrue(math.isclose(lca.score, expected_score, rel_tol=1e-9))
             
    def test_two_consecutive_TD_in_the_past_interpol_db_no_overlap(self):
        ''' 
        Test if two consecutive temporal distributions in the past are correctly aggregated
        TD has been selected to notvmatch time of the background databases directly, so that interpolation between dbs is needed
        Only unitary processes implemented
        
        '''
        
        def prepare_medusa_dbs():
            bd.projects.set_current("__test_tds__")
            
            bd.Database('temporalis-bio').write({
                ('temporalis-bio', "CO2"): {
                    "type": "emission",
                    "name": "carbon dioxide",
                    "temporalis code": "co2",
                },
                ('temporalis-bio', "CH4"): {
                    "type": "emission",
                    "name": "methane",
                    "temporalis code": "ch4",
                },
            })

            bd.Database('background_2022').write({
                ('background_2022', 'C'): {
                'name': 'process C',
                "location": "somewhere",
                'reference product': 'C',
                'exchanges': [
                        {
                            'amount': 1,
                            'type': 'production',
                            'input': ('background_2022', 'C'),
                        },
                        {
                            'amount': 10,
                            'type': 'biosphere',
                            'input': ('temporalis-bio', 'CO2'),
                        },  ]},
            },

                    )

            bd.Database('background_2020').write({
                ('background_2020', 'C'): {
                'name': 'process C',
                "location": "somewhere",
                'reference product': 'C',
                'exchanges': [
                        {
                            'amount': 1,
                            'type': 'production',
                            'input': ('background_2020', 'C'),
                        },
                        {
                            'amount': 30,
                            'type': 'biosphere',
                            'input': ('temporalis-bio', 'CO2'),
                        },  ]},
            },

                    )

            bd.Database('foreground').write({
                ('foreground', 'A'): {
                    'name': 'process A',
                    "location": "somewhere",
                    'reference product': 'A',
                    'exchanges': [
                        {
                            'amount': 1,
                            'type': 'production',
                            'input': ('foreground', 'A'),
                        },
                        {
                            'amount': 1,
                            'type': 'technosphere',
                            'input': ('foreground', 'B'),
                            'temporal_distribution': TemporalDistribution(
                                np.array([-3, -1], dtype='timedelta64[Y]'),
                                np.array([0.5, 0.5])),  
                        },
                     
                       
                    ]
                },
                ('foreground', 'B'):
                {
                    "name": "process B",
                    "location": "somewhere",
                    'reference product': 'B',
                    "exchanges": [
                        {
                            'amount': 1,
                            'type': 'technosphere',
                            'input': ('background_2022', 'C'),
                            'temporal_distribution': TemporalDistribution(
                                np.array([-1, 0], dtype='timedelta64[Y]'), 
                                np.array([0.5, 0.5])), 
                        },
                        {
                            'amount': 1,
                            'type': 'production',
                            'input': ('foreground', 'B'), 
                        }
                    ]

                },
                

    })
            
            
            bd.Method(("GWP", "example")).write([
            (("temporalis-bio", "CO2"), 1),
            ])    
            return 
        
        prepare_medusa_dbs()
        bd.projects.set_current("__test_tds__")
        bd.databases
        
        demand = {('foreground', 'A'): 1}
        method = ("GWP", "example")
        
        # calculate static LCA'
        slca = bc.LCA(demand, method)
        slca.lci()
        slca.lcia()     
        
        # calculate Medusa LCA
        SKIPPABLE = []
        # [node.id for node in bd.Database('background_2022')] + [
        #             node.id for node in bd.Database('background_2020')
        # ]

        def filter_function(database_id: int) -> bool:
            return database_id in SKIPPABLE
        
        eelca = EdgeExtracter(slca, edge_filter_function=filter_function)
        
        timeline = eelca.build_edge_timeline()
        
        database_date_dict = {
                datetime.strptime("2020-01", "%Y-%m"): 'background_2020',
                datetime.strptime("2022-01", "%Y-%m"): 'background_2022',        
            }

        timeline_df = create_grouped_edge_dataframe(timeline, database_date_dict, temporal_grouping= 'year', interpolation_type="linear")

        demand_timing_dict = create_demand_timing_dict(timeline_df, demand)

        dp = create_datapackage_from_edge_timeline(timeline_df, database_date_dict, demand_timing_dict)
        
        fu, data_objs, remapping = prepare_medusa_lca_inputs(demand=demand, demand_timing_dict=demand_timing_dict, method=method) 
        
        lca = bc.LCA(fu, data_objs = data_objs + [dp], remapping_dicts=remapping)
        lca.lci()
        lca.lcia()
        
        #calculate expected score:
        expected_score = 0.25 * 30 + 0.25 * 0.5 *30 + 0.25 * 0.5 * 10 + 0.25 *10 + 0.25 * 10  # 2020 -> 30, 2021 50:50: 30 & 10, 2022 -> 10, 2023 -> 10
        
        print('\nTest test two consecutive TD in the past, interpolation between dbs:')
        print(timeline_df)
        print('Static LCA Score:', slca.score)
        print('MEDUSA LCA Score:', lca.score)
        print('Expected MEDUSA LCA Score:', expected_score)
        
        techno=pd.DataFrame(lca.technosphere_matrix.toarray())
        matrix_ids= [item[1] for item in lca.dicts.activity.reversed.items()]
        techno.index = matrix_ids
        techno.columns = matrix_ids
        techno.to_csv("techno.csv", sep = ';')
        
        for key in lca.activity_dict:
            if key < 2000:
                print(key, "->",bd.get_activity(key)['name'], bd.get_activity(key)["database"]) #BW does not find the "exploded nodes", because they exist only in the datapackages?
        
        print(techno)
        
        # Check if the results are equal using math.isclose
        self.assertTrue(math.isclose(lca.score, expected_score, rel_tol=1e-9))
     
    """    
    def test_temporal_grouping_years(self):
        # create a simple foreground database with 2 processes A and B, where B consumes C from background database
        # C is produced in background database in 2022 and 2024, with different CO2 emissions
        # A consumes B with a temporal distribution of 1/3, 1/3, 1/3
        # The expected result is a medusa LCA score of 12.5 for A
        # The test checks if the monthly aggregation and LCA score is calculated correctly
        
        def prepare_medusa_dbs():
            bd.projects.set_current("__test_medusa__")

            bd.Database('medusa-bio').write({
                        ('medusa-bio', "CO2"): {
                            "type": "biosphere",
                            "name": "carbon dioxide",
                            "temporalis code": "co2",
                        },
                    },
            )
            
            bd.Database("background_2024").write(
                {
                    ("background_2024", "C"): {
                        'name': 'C',
                        'location': 'somewhere',
                        'reference product': 'C',
                        "exchanges": [
                            {
                                'amount': 1,
                                'type': 'production',
                                'input': ("background_2024", 'C'),
                            },
                            
                            {
                                "amount": 10,
                                "input": ("medusa-bio", "CO2"),
                                "type": "biosphere",
                            },
                        ],       
                    },
                },
                
            )
            
            bd.Database("background_2022").write(
                {
                    ("background_2022", "C"): {
                        'name': 'C',
                        'location': 'somewhere',
                        'reference product': 'C',
                        "exchanges": [
                            {
                                'amount': 1,
                                'type': 'production',
                                'input': ("background_2022", 'C'),
                            },                   
                            {
                                "amount": 15,
                                "input": ("medusa-bio", "CO2"),
                                "type": "biosphere",
                            },
                        ],   
                    },
                },
                
            )
            
            bd.Database("foreground").write(
                {
                    ("foreground", "A"): {
                        'name': 'A',
                        'location': 'somewhere',
                        'reference product': 'A',
                        "exchanges": [
                            {
                                'amount': 1,
                                'type': 'production',
                                'input': ('foreground', 'A'),
                            },                 
                            {
                                "amount": 1,
                                "input": ("foreground", "B"),
                                'temporal_distribution': TemporalDistribution(
                                        np.array([-36, -24, -12, 0], dtype='timedelta64[M]'),
                                        np.array([1/4, 1/4, 1/4, 1/4])),  
                                "type": "technosphere",
                            },
                        ],
                    },
                    
                    ("foreground", "B"): {
                        'name': 'B',
                        'location': 'somewhere',
                        'reference product': 'B',
                        "exchanges": [
                            {
                                'amount': 1,
                                'type': 'production',
                                'input': ('foreground', 'B'),
                            },
                            {
                                "amount": 1,
                                "input": ("background_2024", "C"),
                                "type": "technosphere",
                            },
                        ],
                    },
                }
            )
        
            bd.Method(("GWP", "example")).write([
            (("medusa-bio", "CO2"), 1),
            ])    
            return 
        
        prepare_medusa_dbs()
        bd.projects.set_current("__test_medusa__") # monthly grouping
        bd.databases
        
        demand = {('foreground', 'A'): 1}
        method = ("GWP", "example")
        
        # calculate static LCA'
        slca = bc.LCA(demand, method)
        slca.lci()
        slca.lcia()       
        
        # calculate Medusa LCA
        SKIPPABLE = [node.id for node in bd.Database('background_2022')] + [
                    node.id for node in bd.Database('background_2024')
        ]

        def filter_function(database_id: int) -> bool:
            return database_id in SKIPPABLE
        
        eelca = EdgeExtracter(slca, edge_filter_function=filter_function)
        
        timeline = eelca.build_edge_timeline()
        
        database_date_dict = {
                datetime.strptime("2022-01", "%Y-%m"): 'background_2022',
                datetime.strptime("2024-01", "%Y-%m"): 'background_2024',        
            }

        timeline_df = create_grouped_edge_dataframe(timeline, database_date_dict, temporal_grouping= 'year', interpolation_type="linear") # yearly grouping

        demand_timing_dict = create_demand_timing_dict(timeline_df, demand)

        dp = create_datapackage_from_edge_timeline(timeline_df, database_date_dict, demand_timing_dict)
        
        fu, data_objs, remapping = prepare_medusa_lca_inputs(demand=demand, demand_timing_dict=demand_timing_dict, method=method) 
        
        lca = bc.LCA(fu, data_objs = data_objs + [dp], remapping_dicts=remapping)
        lca.lci()
        lca.lcia()
        
        #calculate expected score:
        expected_score = 1/4*15 + 1/4*15 +1/8*15 + 1/8*10 + 1/4*10
        
        print('\nTest Temporal Grouping Years:')     
        print(timeline_df)
        print('Static LCA Score:', slca.score)
        print('MEDUSA LCA Score:', lca.score)
        print('Expected MEDUSA LCA Score:', expected_score)
        
        
        # Check if the results are equal using math.isclose
        self.assertTrue(math.isclose(lca.score, expected_score, rel_tol=1e-9))
        
        """

if __name__ == '__main__':
    unittest.main()

     

    




class TimeMappingDict(dict):
    """
    This class is used to create a dictionary that maps a tuple of (flow and timestamp) to an unique integer id.
    """

    def __init__(self, start_id=2, *args, **kwargs) -> None:
        """"
        Initializes the `TimeMappingDict` object.

        Parameters
        ----------
        start_id : int, optional
            The starting id for the mapping. Default is 2.
        *args : Variable length argument list
        **kwargs : Arbitrary keyword arguments

        Returns
        -------
        None
        """

        super().__init__(*args, **kwargs)
        self._current_id = start_id
        self._check_id = (
            start_id - 1
        )  # check_id that is different from the start id for the reversed dict

    def add(self, process_time_tuple, unique_id=None):
        """"
        Adds a new process_time_tuple to the `TimeMappingDict` object.

        Parameters
        ----------
        process_time_tuple : tuple
            A tuple of (flow and timestamp)
        unique_id : int, optional
            An unique id for the process_time_tuple. Default is None.

        Returns
        -------
        int
            An unique id for the process_time_tuple, and adds it to the dictionary, if not already present.
        """
        if process_time_tuple in self:
            return self[process_time_tuple]

        if unique_id is not None:
            self[process_time_tuple] = unique_id
            return unique_id

        self[process_time_tuple] = self._current_id
        self._current_id += 1

        return self._current_id - 1

    def reversed(self):
        """return a reversed version of dict, update if necessary

        Parameters
        ----------
        None

        Returns
        -------
        dict
            A reversed dictionary of the `TimeMappingDict` object.
        """


        if self._check_id != self._current_id:
            self.reversed_dict = {v: k for k, v in self.items()}
            self._check_id = self._current_id
        return self.reversed_dict



"""
Testing the results of the static LCA and the MedusaLCA to see if the new interpolated amounts are correct.
"""

import math
import unittest
from bw_temporalis import easy_timedelta_distribution, TemporalDistribution
from edge_extractor import EdgeExtracter
from medusa_tools import *
import bw2data as bd
import bw2calc as bc
import numpy as np


class TestAmounts(unittest.TestCase):
    def test_amounts(self):
        bd.projects.set_current("__test_medusa__")
        bd.Database("temporalis-bio").write(
            {
                ("temporalis-bio", "CO2"): {
                    "type": "emission",
                    "name": "carbon dioxide",
                    "temporalis code": "co2",
                },
            }
        )

        bd.Database("background_2024").write(
            {
                ("background_2024", "electricity_mix"): {
                    "name": "Electricity mix",
                    "location": "somewhere",
                    "reference product": "electricity mix",
                    "exchanges": [
                        {
                            "amount": 1,
                            "type": "production",
                            "input": ("background_2024", "electricity_mix"),
                        },
                        {
                            "amount": 1,
                            "type": "technosphere",
                            "input": ("background_2024", "electricity_wind"),
                        },
                    ],
                },
                ("background_2024", "electricity_wind"): {
                    "name": "Electricity production, wind",
                    "location": "somewhere",
                    "reference product": "electricity, wind",
                    "exchanges": [
                        {
                            "amount": 1,
                            "type": "production",
                            "input": ("background_2024", "electricity_wind"),
                        },
                        {
                            "amount": 0.9,
                            "type": "biosphere",
                            "input": ("temporalis-bio", "CO2"),
                        },
                    ],
                },
            }
        )

        bd.Database("background_2020").write(
            {
                ("background_2020", "electricity_mix"): {
                    "name": "Electricity mix",
                    "location": "somewhere",
                    "reference product": "electricity mix",
                    "exchanges": [
                        {
                            "amount": 1,
                            "type": "production",
                            "input": ("background_2020", "electricity_mix"),
                        },
                        {
                            "amount": 1,
                            "type": "technosphere",
                            "input": ("background_2020", "electricity_wind"),
                        },
                    ],
                },
                ("background_2020", "electricity_wind"): {
                    "name": "Electricity production, wind",
                    "location": "somewhere",
                    "reference product": "electricity, wind",
                    "exchanges": [
                        {
                            "amount": 1,
                            "type": "production",
                            "input": ("background_2020", "electricity_wind"),
                        },
                        {
                            "amount": 1.2,
                            "type": "biosphere",
                            "input": ("temporalis-bio", "CO2"),
                        },
                    ],
                },
            }
        )

        bd.Database("foreground").write(
            {
                ("foreground", "electrolysis"): {
                    "name": "Hydrogen production, electrolysis",
                    "location": "somewhere",
                    "reference product": "hydrogen",
                    "exchanges": [
                        {
                            "amount": 1,
                            "type": "production",
                            "input": ("foreground", "electrolysis"),
                        },
                        {
                            "amount": 5,
                            "type": "technosphere",
                            "input": ("background_2024", "electricity_mix"),
                            "temporal_distribution": easy_timedelta_distribution(  # e.g. because some hydrogen was stored in the meantime
                                start=-1,
                                end=0,  # Range includes both start and end
                                resolution="Y",  # M for months, Y for years, etc.
                                steps=2,
                            ),
                        },
                    ],
                },
                ("foreground", "heat_from_hydrogen"): {
                    "name": "Heat production, hydrogen",
                    "location": "somewhere",
                    "reference product": "heat",
                    "exchanges": [
                        {
                            "amount": 1,
                            "type": "production",
                            "input": ("foreground", "heat_from_hydrogen"),
                        },
                        {
                            "amount": 0.7,
                            "type": "technosphere",
                            "input": ("foreground", "electrolysis"),
                            "temporal_distribution": easy_timedelta_distribution(  # e.g. because some hydrogen was stored in the meantime
                                start=-2,
                                end=0,  # Range includes both start and end
                                resolution="Y",  # M for months, Y for years, etc.
                                steps=2,
                            ),
                        },
                    ],
                },
            }
        )

        bd.Method(("GWP", "example")).write(
            [
                (("temporalis-bio", "CO2"), 1),
            ]
        )

        demand = {("foreground", "heat_from_hydrogen"): 1}
        gwp = ("GWP", "example")

        # calculate static LCA'
        slca = bc.LCA(demand, gwp)
        slca.lci()
        slca.lcia()

        # calculate Medusa LCA
        SKIPPABLE = [node.id for node in bd.Database("background_2020")] + [
            node.id for node in bd.Database("background_2024")
        ]

        def filter_function(database_id: int) -> bool:
            return database_id in SKIPPABLE

        eelca = EdgeExtracter(slca, edge_filter_function=filter_function)
        timeline = eelca.build_edge_timeline()

        database_date_dict = {
            datetime.strptime("2020", "%Y"): "background_2020",
            datetime.strptime("2024", "%Y"): "background_2024",
        }

        timeline_df = create_grouped_edge_dataframe(
            timeline, database_date_dict, interpolation_type="linear"
        )

        demand_timing_dict = create_demand_timing_dict(timeline_df, demand)

        dp = create_datapackage_from_edge_timeline(
            timeline_df, database_date_dict, demand_timing_dict
        )

        fu, data_objs, remapping = prepare_medusa_lca_inputs(
            demand=demand, demand_timing_dict=demand_timing_dict, method=gwp
        )
        lca = bc.LCA(fu, data_objs=data_objs + [dp], remapping_dicts=remapping)
        lca.lci()
        lca.lcia()

        # calculate expected score:
        expected_score = 3.4125

        print("Expected score: ", expected_score)
        print("Static LCA score: ", slca.score)
        print("Medusa LCA score: ", lca.score)
        # Check if the results are equal using math.isclose
        self.assertTrue(math.isclose(lca.score, expected_score, rel_tol=1e-1))


if __name__ == "__main__":
    unittest.main()



"""
Testing the results of the static LCA and the MedusaLCA to see if the new interpolated amounts are correct.
"""

import math
import unittest
import bw2data as bd
import bw2calc as bc
import numpy as np
from timex_lca import MedusaLCA
from tests.databases import db_abc_loopA_with_biosphere_tds_CO2_and_CH4
from datetime import datetime


class TestBioflows(unittest.TestCase):

    def test_advanced_bioflows(self):
        import warnings

        warnings.filterwarnings("ignore")

        db_abc_loopA_with_biosphere_tds_CO2_and_CH4()
        database_date_dict = {
            "background_2008": datetime.strptime("2008", "%Y"),
            "background_2024": datetime.strptime("2024", "%Y"),
            "foreground": "dynamic",  # flag databases that should be temporally distributed with "dynamic"
        }
        SKIPPABLE = [node.id for node in bd.Database("background_2008")] + [
            node.id for node in bd.Database("background_2024")
        ]

        def filter_function(database_id: int) -> bool:
            return database_id in SKIPPABLE

        demand = {("foreground", "A"): 1}
        method = ("GWP", "example")

        mlca = MedusaLCA(
            demand, method, filter_function, database_date_dict, max_calc=5000
        )
        mlca.build_timeline()
        mlca.build_datapackage()
        mlca.lci()
        mlca.lcia()
        self.assertTrue(
            math.isclose(mlca.score, mlca.static_lca.score, rel_tol=1e-2),
            f"Total scores didn't match up. Medusa LCA score is {mlca.score}, static LCA score is {mlca.static_lca.score}",
        )

        mlca.build_dynamic_biosphere()
        mlca.calculate_dynamic_biosphere_lci()

        for idx, row in enumerate(mlca.lca.inventory.toarray()):
            bioflow_matrix_id = mlca.lca.dicts.biosphere.reversed[idx]
            bioflow_name = bd.get_node(id=bioflow_matrix_id)["code"]
            self.assertTrue(
                math.isclose(
                    row.sum(),
                    mlca.dynamic_inventory[bioflow_name]["amount"].sum(),
                    rel_tol=1e-7,
                ),
                f"Something didn't match. lca.inventory says {row.sum()} and dynamic_inventory says {mlca.dynamic_inventory[bioflow_name]['amount'].sum()}",
            )


if __name__ == "__main__":
    unittest.main()



from .timex_lca import TimexLCA
from .matrix_modifier import MatrixModifier
from .edge_extractor import EdgeExtractor
from .timeline_builder import TimelineBuilder
from .dynamic_biosphere_builder import DynamicBiosphereBuilder
from .dynamic_characterization import DynamicCharacterization

__version__ = '0.1.1' 


import pandas as pd
import bw2data as bd
import numpy as np
import os
import warnings
import json

from typing import Union, Tuple, Optional, Callable, List
from datetime import datetime


class DynamicCharacterization:
    """
    This class calculates dynamic characterization of life cycle inventories with temporal information.

    """

    def __init__(
        self,
        dynamic_inventory_df: pd.DataFrame,
        activity_dict: dict,
        biosphere_dict: dict,
        activity_time_mapping_dict_reversed: dict,
        biosphere_time_mapping_dict_reversed: dict,
        demand_timing_dict: dict,
        temporal_grouping: dict,
        method: tuple,
        characterization_function_dict: dict = None,
    ):
        """
        Initializes the DynamicCharacterization object. In case users don't provide own dynamic characterization functions, it adds dynamic characterization functions for the GHGs mentioned in the
        IPCC AR6 Chapter 7, if these GHG are also characterized in the selected static LCA method.

        Parameters
        ----------
        dynamic_inventory_df : pd.DataFrame
            Dynamic inventory, formatted as a DataFrame, which contains the timing, id and amount of emissions and the emitting activity.
        activity_dict : dict
            Dictionary mapping activity ids to their matrix index
        biosphere_dict : dict
            Dictionary mapping biosphere flow ids to their matrix index
        activity_time_mapping_dict_reversed : dict
            Reversed `activity_time_mapping_dict`: {time_mapping_id: ((('database', 'code'), datetime_as_integer)} 
        biosphere_time_mapping_dict_reversed : dict
            Reversed `biosphere_time_mapping_dict`: {time_mapping_id: ((('database', 'code'), datetime_as_integer)} 
        demand_timing_dict : dict
            A dictionary mapping the demand(s) to its timing
        temporal_grouping : str
            A string indicating the temporal grouping of the processes, e.g. 'year', 'month', 'day', 'hour'
        method : tuple
            Tuple of the selcted the LCIA method, e.g. `("EF v3.1", "climate change", "global warming potential (GWP100)")`
        characterization_function_dict : dict, optional
            A dictionary mapping biosphere flow ids to user-provided dynamic characterization functions, by default None.

        Returns
        -------
        None
        
        """
        self.dynamic_inventory_df = dynamic_inventory_df
        self.activity_dict = activity_dict
        self.biosphere_dict = biosphere_dict
        self.activity_time_mapping_dict_reversed = activity_time_mapping_dict_reversed
        self.biosphere_time_mapping_dict_reversed = biosphere_time_mapping_dict_reversed
        self.demand_timing_dict = demand_timing_dict
        self.temporal_grouping = temporal_grouping
        self.method = method

        if not characterization_function_dict:
            warnings.warn(
                f"No custom dynamic characterization functions provided. Using default dynamic characterization functions based on IPCC AR6 meant to work with biosphere3 flows. The flows that are characterized are based on the selection of the initially chosen impact category: {self.method}. You can look up the mapping in the timex_lca.dynamic_characterizer.characterization_function_dict."
            )
            self.add_default_characterization_functions()

        else:
            self.characterization_function_dict = characterization_function_dict  # user-provided characterization functions

    def characterize_dynamic_inventory(
        self,
        metric: str | None = "GWP",  # available metrics are "radiative_forcing" and "GWP", defaulting to GWP
        time_horizon: int | None = 100,
        fixed_time_horizon: (
            bool | None
        ) = False, 
        cumsum: bool | None = True,
    ) -> Tuple[pd.DataFrame, str, bool, int]:
        """
        Characterizes the dynamic inventory, formatted as a Dataframe, by evaluating each emission (row in DataFrame) using given dynamic characterization functions.

        Available metrics are radiative forcing [W/m2] and GWP [kg CO2eq], defaulting to `radiative_forcing`.

        The `characterization_functions` are already created during __init__ and stored in the dictionary `DynamicCharacterization.characterization_function_dict`. 
        In this method, they are applied to each row of the timeline-DataFrame for the duration of `time_horizon`, defaulting to 100 years.
        The `fixed_time_horizon` parameter determines whether the evaluation time horizon for all emissions is calculated from the 
        functional unit (`fixed_time_horizon=True`), regardless of when the actual emission occurs, or from the time of the emission itself(`fixed_time_horizon=False`). 
        The former is the implementation of the Levasseur approach (https://doi.org/10.1021/es9030003), while the latter is how conventional LCA is done.
        The Levasseur approach means that earlier emissions are characterized for a longer time period than later emissions.

        Parameters
        ----------
        metric : str, optional
            the metric for which the dynamic LCIA should be calculated. Default is "GWP". Available: "GWP" and "radiative_forcing"
        time_horizon: int, optional
            the time horizon for the dynamic characterization. Default is 100 years.
        fixed_time_horizon: bool, optional
            If True, the time horizon is calculated from the time of the functional unit (FU) instead of the time of emission. Default is False.
        cumsum: bool, optional
            whether to calculate the cumulative sum of the characterization results. Default is True.

        Returns
        -------
        pd.DataFrame
            characterized dynamic inventory
        """

        if metric not in {"radiative_forcing", "GWP"}:
            raise ValueError(
                f"Metric must be either 'radiative_forcing' or 'GWP', not {metric}"
            )
        
        if metric == "GWP":
            warnings.warn(
                        "Using timex_lca's default CO2 characterization function for GWP reference."
                    )

        time_res_dict = {
            "year": "%Y",
            "month": "%Y%m",
            "day": "%Y%m%d",
            "hour": "%Y%m%d%M",
        }

        self.characterized_inventory = pd.DataFrame()

        for _, row in self.dynamic_inventory_df.iterrows():

            if row.flow not in self.characterization_function_dict.keys(): #skip uncharacterized biosphere flows
                continue

            if metric == "radiative_forcing":  # radiative forcing in W/m2

                if (
                    not fixed_time_horizon
                ):  # fixed_time_horizon = False: conventional approach, emission is calculated from t emission for the length of time horizon
                    self.characterized_inventory = pd.concat(
                        [
                            self.characterized_inventory,
                            self.characterization_function_dict[row.flow]( #here the dynamic characterization function is called and applied to the emission of the row
                                row,
                                period=time_horizon,
                            ),
                        ]
                    )

                else:  # fixed_time_horizon = True: Levasseur approach: time_horizon for all emissions starts at timing of FU + time_horizon
                    # e.g. an emission occuring n years before FU is characterized for time_horizon+n years
                    timing_FU = [value for value in self.demand_timing_dict.values()]  
                    end_TH_FU_list = [x + time_horizon for x in timing_FU]

                    if len(end_TH_FU_list) > 1:
                        warnings.warn(
                            f"There are multiple functional units with different timings. The first one ({str(end_TH_FU_list[0])}) will be used as a basis for the fixed time horizon in dynamic characterization."
                        )

                    end_TH_FU = datetime.strptime(
                        str(end_TH_FU_list[0]), time_res_dict[self.temporal_grouping]
                    )

                    timing_emission = (
                        row.date.to_pydatetime()
                    )  # convert 'pandas._libs.tslibs.timestamps.Timestamp' to datetime object
                    new_TH = round(
                        (end_TH_FU - timing_emission).days / 365.25
                    )  # time difference in integer years between emission timing and end of time horizon of FU

                    self.characterized_inventory = pd.concat(
                        [
                            self.characterized_inventory,
                            self.characterization_function_dict[row.flow](
                                row,
                                period=new_TH,
                            ),
                        ]
                    )

            if metric == "GWP":  # scale radiative forcing to GWP [kg CO2 equivalent]
                if (
                    not fixed_time_horizon
                ):  # fixed_time_horizon = False: conventional approach, emission is calculated from t emission for the length of time_horizon
                    
                    radiative_forcing_ghg = self.characterization_function_dict[
                        row.flow
                    ](
                        row,
                        period=time_horizon,
                    )

                    row["amount"] = 1  # convert 1 kg CO2 equ.
                    radiative_forcing_co2 = characterize_co2(row, period=time_horizon)

                    ghg_integral = radiative_forcing_ghg["amount"].sum()
                    co2_integral = radiative_forcing_co2["amount"].sum()
                    co2_equiv = ghg_integral / co2_integral

                    row_data = {
                        "date": radiative_forcing_ghg.loc[
                            0, "date"
                        ],  # start date of emission
                        "amount": co2_equiv,  # ghg emission in kg CO2-equ
                        "flow": radiative_forcing_ghg.loc[0, "flow"],
                        "activity": radiative_forcing_ghg.loc[0, "activity"],
                    }

                    row_df = pd.DataFrame([row_data])
                    self.characterized_inventory = pd.concat(
                        [self.characterized_inventory, row_df], ignore_index=True
                    )

                else:  # fixed_time_horizon = True: Levasseur approach: time_horizon for all emissions starts at timing of FU + time_horizon
                    # e.g. an emission occuring n years before FU is characterized for time_horizon+n years
                    timing_FU = [
                        value for value in self.demand_timing_dict.values()
                    ] 
                    end_TH_FU_list = [x + time_horizon for x in timing_FU]

                    if len(end_TH_FU_list) > 1:
                        warnings.warn(
                            f"There are multiple functional units with different timings. The first one ({str(end_TH_FU_list[0])}) will be used as a basis for the fixed time horizon in dynamic characterization."
                        )

                    end_TH_FU = datetime.strptime(
                        str(end_TH_FU_list[0]), time_res_dict[self.temporal_grouping]
                    )

                    timing_emission = (
                        row.date.to_pydatetime()
                    )  
                    new_TH = round(
                        (end_TH_FU - timing_emission).days / 365.25
                    )  # time difference in integer years between emission timing and end of TH of FU

                    radiative_forcing_ghg = self.characterization_function_dict[
                        row.flow
                    ](
                        row,
                        period=new_TH,
                    )  # indidvidual emissions are calculated for t_emission until t_FU + time_horizon

                    row["amount"] = 1  # convert 1 kg CO2 equ.
                    radiative_forcing_co2 = characterize_co2(
                        row, period=time_horizon
                    )  # reference substance CO2 is calculated for length of time horizon!

                    ghg_integral = radiative_forcing_ghg["amount"].sum()
                    co2_integral = radiative_forcing_co2["amount"].sum()
                    co2_equiv = ghg_integral / co2_integral

                    row_data = {
                        "date": radiative_forcing_ghg.loc[
                            0, "date"
                        ],  # start date of emission
                        "amount": co2_equiv,  # ghg emission in CO2 equiv
                        "flow": radiative_forcing_ghg.loc[0, "flow"],
                        "activity": radiative_forcing_ghg.loc[0, "activity"],
                    }
                    row_df = pd.DataFrame([row_data])
                    self.characterized_inventory = pd.concat(
                        [self.characterized_inventory, row_df], ignore_index=True
                    )

            # sort by date
            if "date" in self.characterized_inventory:
                self.characterized_inventory.sort_values(
                    by="date", ascending=True, inplace=True
                )
                self.characterized_inventory.reset_index(drop=True, inplace=True)

            if cumsum and "amount" in self.characterized_inventory:
                self.characterized_inventory["amount_sum"] = (
                    self.characterized_inventory["amount"].cumsum()
                )  # TODO: there is also an option for cumulative results in the characterization functions themselves. Rethink where this is handled best and to avoid double cumsum

           
        if self.characterized_inventory.empty:
            raise ValueError(
                "No flows were characterized. Check if the flows in the dynamic inventory are available the dynamic characterization functions."
            )

        # remove rows with zero amounts to make it more readable
        self.characterized_inventory = self.characterized_inventory[
            self.characterized_inventory["amount"] != 0
        ]

        # add meta data and reorder
        self.characterized_inventory["activity_name"] = self.characterized_inventory[
            "activity"
        ].map(lambda x: self.activity_time_mapping_dict_reversed.get(x)[0])

        self.characterized_inventory["flow_name"] = self.characterized_inventory[
            "flow"
        ].apply(lambda x: bd.get_node(id=x)["name"])

        self.characterized_inventory = self.characterized_inventory[
            [
                "date",
                "amount",
                "flow",
                "flow_name",
                "activity",
                "activity_name",
                "amount_sum",
            ]
        ]
        self.characterized_inventory.reset_index(drop=True, inplace=True)
        self.characterized_inventory.sort_values(
            by="date", ascending=True, inplace=True
        )

        return self.characterized_inventory

    def add_default_characterization_functions(self):
        """
        Add default dynamic characterization functions for CO2, CH4, N2O and other GHGs, based on IPCC AR6 Chapter 7 decay curves.

        Please note: Currently, only CO2, CH4 and N2O include climate-carbon feedbacks. 
        
        This has not yet been added for other GHGs. Refer to https://esd.copernicus.org/articles/8/235/2017/esd-8-235-2017.html"
        "Methane, non-fossil" is currently also excluded from the default characterization functions, as it has a different static CF than fossil methane and we need to check the correct value (#TODO)

        Parameters
        ----------
        None

        Returns
        -------
        None but adds default dynamic characterization functions to the `characterization_function_dict` attribute of the DynamicCharacterization object.

        """
        self.characterization_function_dict = dict()

        bioflows_in_lcia_method = bd.Method(self.method).load()

        filepath = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "data", "decay_multipliers.json"
        )
        
        with open(filepath) as json_file:
            decay_multipliers = json.load(json_file)

        for flow in bioflows_in_lcia_method:
            node = bd.get_node(database=flow[0][0], code=flow[0][1])

            if "carbon dioxide" in node["name"].lower():
                if "soil" in node.get("categories", []):
                    self.characterization_function_dict[node.id] = (
                        characterize_co2_uptake  # negative emission because uptake by soil
                    )

                else:
                    self.characterization_function_dict[node.id] = characterize_co2

            elif (
                "methane, fossil" in node["name"].lower()
                or "methane, from soil or biomass stock" in node["name"].lower()
            ):
                # TODO Check why "methane, non-fossil" has a CF of 27 instead of 29.8, currently excluded
                self.characterization_function_dict[node.id] = characterize_ch4

            elif "dinitrogen monoxide" in node["name"].lower():
                self.characterization_function_dict[node.id] = characterize_n2o

            elif "carbon monoxide" in node["name"].lower():
                self.characterization_function_dict[node.id] = characterize_co

            else:
                cas_number = node.get("CAS number")
                if cas_number:
                    decay_series = decay_multipliers.get(cas_number)
                    if decay_series is not None:
                        self.characterization_function_dict[node.id] = (
                            create_generic_characterization_function(np.array(decay_series))
                        )


def IRF_co2(year) -> callable:
    """
    Impulse Resonse Function (IRF) of CO2

    Parameters
    ----------
    year : int
        The year after emission for which the IRF is calculated.
    
    Returns
    -------
    float
        The IRF value for the given year.

    """
    alpha_0, alpha_1, alpha_2, alpha_3 = 0.2173, 0.2240, 0.2824, 0.2763
    tau_1, tau_2, tau_3 = 394.4, 36.54, 4.304
    exponentials = lambda year, alpha, tau: alpha * tau * (1 - np.exp(-year / tau))
    return (
        alpha_0 * year
        + exponentials(year, alpha_1, tau_1)
        + exponentials(year, alpha_2, tau_2)
        + exponentials(year, alpha_3, tau_3)
    )


def characterize_co2(
    series,
    period: int | None = 100,
    cumulative: bool | None = False,
) -> pd.DataFrame:
    """
    Calculate the cumulative or marginal radiative forcing (CRF) from CO2 for each year in a given period.
    
    Based on characterize_co2 from bw_temporalis, but updated numerical values from IPCC AR6 Ch7 & SM.

    If `cumulative` is True, the cumulative CRF is calculated. If `cumulative` is False, the marginal CRF is calculated.
    Takes a single row of the TimeSeries Pandas DataFrame (corresponding to a set of (`date`/`amount`/`flow`/`activity`).
    For each year in the given period, the CRF is calculated.
    Units are watts/square meter/kilogram of CO2.

    Returns
    -------
    A TimeSeries dataframe with the following columns:
    - date: datetime64[s]
    - amount: float
    - flow: str
    - activity: str

    See also
    --------
    Joos2013: Relevant scientific publication on CRF: https://doi.org/10.5194/acp-13-2793-2013
    Schivley2015: Relevant scientific publication on the numerical calculation of CRF: https://doi.org/10.1021/acs.est.5b01118
    Forster2023: Updated numerical values from IPCC AR6 Chapter 7 (Table 7.15): https://doi.org/10.1017/9781009157896.009
    """

    # functional variables and units (from publications listed in docstring)
    radiative_efficiency_ppb = (
        1.33e-5  # W/m2/ppb; 2019 background co2 concentration; IPCC AR6 Table 7.15
    )

    # for conversion from ppb to kg-CO2
    M_co2 = 44.01  # g/mol
    M_air = 28.97  # g/mol, dry air
    m_atmosphere = 5.135e18  # kg [Trenberth and Smith, 2005]

    radiative_efficiency_kg = (
        radiative_efficiency_ppb * M_air / M_co2 * 1e9 / m_atmosphere
    )  # W/m2/kg-CO2

    date_beginning: np.datetime64 = series["date"].to_numpy()
    date_characterized: np.ndarray = date_beginning + np.arange(
        start=0, stop=period, dtype="timedelta64[Y]"
    ).astype("timedelta64[s]")

    decay_multipliers: np.ndarray = np.array(
        [radiative_efficiency_kg * IRF_co2(year) for year in range(period)]
    )

    forcing = pd.Series(data=series.amount * decay_multipliers, dtype="float64")

    if not cumulative:
        forcing = forcing.diff(periods=1).fillna(0)

    return pd.DataFrame(
        {
            "date": pd.Series(data=date_characterized, dtype="datetime64[s]"),
            "amount": forcing,
            "flow": series.flow,
            "activity": series.activity,
        }
    )


def characterize_co2_uptake(
    series,
    period: int | None = 100,
    cumulative: bool | None = False,
) -> pd.DataFrame:
    """
    The same as characterize_co2, but with a negative sign for uptake of CO2.

    Based on characterize_co2 from bw_temporalis, but updated numerical values from IPCC AR6 Ch7 & SM.

    Calculate the negative cumulative or marginal radiative forcing (CRF) from CO2-uptake for each year in a given period.

    If `cumulative` is True, the cumulative CRF is calculated. If `cumulative` is False, the marginal CRF is calculated.
    Takes a single row of the TimeSeries Pandas DataFrame (corresponding to a set of (`date`/`amount`/`flow`/`activity`).
    For each year in the given period, the CRF is calculated.
    Units are watts/square meter/kilogram of CO2.

    Returns
    -------
    A TimeSeries dataframe with the following columns:
    - date: datetime64[s]
    - amount: float
    - flow: str
    - activity: str

    See also
    --------
    Joos2013: Relevant scientific publication on CRF: https://doi.org/10.5194/acp-13-2793-2013
    Schivley2015: Relevant scientific publication on the numerical calculation of CRF: https://doi.org/10.1021/acs.est.5b01118
    Forster2023: Updated numerical values from IPCC AR6 Chapter 7 (Table 7.15): https://doi.org/10.1017/9781009157896.009
    """

    # functional variables and units (from publications listed in docstring)
    radiative_efficiency_ppb = (
        1.33e-5  # W/m2/ppb; 2019 background co2 concentration; IPCC AR6 Table 7.15
    )

    # for conversion from ppb to kg-CO2
    M_co2 = 44.01  # g/mol
    M_air = 28.97  # g/mol, dry air
    m_atmosphere = 5.135e18  # kg [Trenberth and Smith, 2005]

    radiative_efficiency_kg = (
        radiative_efficiency_ppb * M_air / M_co2 * 1e9 / m_atmosphere
    )  # W/m2/kg-CO2

    date_beginning: np.datetime64 = series["date"].to_numpy()
    date_characterized: np.ndarray = date_beginning + np.arange(
        start=0, stop=period, dtype="timedelta64[Y]"
    ).astype("timedelta64[s]")

    decay_multipliers: np.ndarray = np.array(
        [radiative_efficiency_kg * IRF_co2(year) for year in range(period)]
    )

    forcing = pd.Series(data=series.amount * decay_multipliers, dtype="float64")

    forcing = (
        -forcing
    )  # flip the sign of the characterization function for CO2 uptake and not release

    if not cumulative:
        forcing = forcing.diff(periods=1).fillna(0)

    return pd.DataFrame(
        {
            "date": pd.Series(data=date_characterized, dtype="datetime64[s]"),
            "amount": forcing,
            "flow": series.flow,
            "activity": series.activity,
        }
    )


def characterize_co(
    series,
    period: int | None = 100,
    cumulative: bool | None = False,
) -> pd.DataFrame:
    """
    Calculate the cumulative or marginal radiative forcing (CRF) from CO for each year in a given period.

    This is exactly the same function as for CO2, it's just scaled by the ratio of molar masses of CO and CO2. This is because CO is very short-lived (lifetime ~2 months) and we assume that it completely reacts to CO2 within the first year.

    Based on characterize_co2 from bw_temporalis, but updated numerical values from IPCC AR6 Ch7 & SM.

    Calculate the cumulative or marginal radiative forcing (CRF) from CO2 for each year in a given period.

    If `cumulative` is True, the cumulative CRF is calculated. If `cumulative` is False, the marginal CRF is calculated.
    Takes a single row of the TimeSeries Pandas DataFrame (corresponding to a set of (`date`/`amount`/`flow`/`activity`).
    For each year in the given period, the CRF is calculated.
    Units are watts/square meter/kilogram of CO2.

    Returns
    -------
    A TimeSeries dataframe with the following columns:
    - date: datetime64[s]
    - amount: float
    - flow: str
    - activity: str

    See also
    --------
    Joos2013: Relevant scientific publication on CRF: https://doi.org/10.5194/acp-13-2793-2013
    Schivley2015: Relevant scientific publication on the numerical calculation of CRF: https://doi.org/10.1021/acs.est.5b01118
    Forster2023: Updated numerical values from IPCC AR6 Chapter 7 (Table 7.15): https://doi.org/10.1017/9781009157896.009
    """

    # functional variables and units (from publications listed in docstring)
    radiative_efficiency_ppb = (
        1.33e-5  # W/m2/ppb; 2019 background co2 concentration; IPCC AR6 Table 7.15
    )

    # for conversion from ppb to kg-CO2
    M_co2 = 44.01  # g/mol
    M_co = 28.01   # g/mol
    M_air = 28.97  # g/mol, dry air
    m_atmosphere = 5.135e18  # kg [Trenberth and Smith, 2005]

    radiative_efficiency_kg = (
        radiative_efficiency_ppb * M_air / M_co2 * 1e9 / m_atmosphere
    )  # W/m2/kg-CO2

    date_beginning: np.datetime64 = series["date"].to_numpy()
    date_characterized: np.ndarray = date_beginning + np.arange(
        start=0, stop=period, dtype="timedelta64[Y]"
    ).astype("timedelta64[s]")

    decay_multipliers: np.ndarray = np.array(
        [
            M_co2 / M_co * radiative_efficiency_kg * IRF_co2(year)
            for year in range(period)
        ]  # <-- Scaling from co2 to co is done here
    )

    forcing = pd.Series(data=series.amount * decay_multipliers, dtype="float64")

    if not cumulative:
        forcing = forcing.diff(periods=1).fillna(0)

    return pd.DataFrame(
        {
            "date": pd.Series(data=date_characterized, dtype="datetime64[s]"),
            "amount": forcing,
            "flow": series.flow,
            "activity": series.activity,
        }
    )


def characterize_ch4(
    series,
    period: int = 100,
    cumulative=False,
) -> pd.DataFrame:
    """
    Calculate the cumulative or marginal radiative forcing (CRF) from CH4 for each year in a given period. 

    Based on characterize_methane from bw_temporalis, but updated numerical values from IPCC AR6 Ch7 & SM.
    
    This DOES include indirect effects of CH4 on ozone and water vapor, but DOES NOT include the decay to CO2. 
    For more info on that, see the deprecated version of bw_temporalis.

    If `cumulative` is True, the cumulative CRF is calculated. If `cumulative` is False, the marginal CRF is calculated.
    Takes a single row of the TimeSeries Pandas DataFrame (corresponding to a set of (`date`/`amount`/`flow`/`activity`).
    For earch year in the given period, the CRF is calculated.
    Units are watts/square meter/kilogram of CH4.

    Parameters
    ----------
    series : array-like
        A single row of the TimeSeries dataframe.
    period : int, optional
        Time period for calculation (number of years), by default 100
    cumulative : bool, optional
        Should the RF amounts be summed over time?

    Returns
    -------
    A TimeSeries dataframe with the following columns:
    - date: datetime64[s]
    - amount: float
    - flow: str
    - activity: str

    See also
    --------
    Joos2013: Relevant scientific publication on CRF: https://doi.org/10.5194/acp-13-2793-2013
    Schivley2015: Relevant scientific publication on the numerical calculation of CRF: https://doi.org/10.1021/acs.est.5b01118
    Forster2023: Updated numerical values from IPCC AR6 Chapter 7 (Table 7.15): https://doi.org/10.1017/9781009157896.009
    """

    # functional variables and units (from publications listed in docstring)
    radiative_efficiency_ppb = 5.7e-4  # # W/m2/ppb; 2019 background cch4 concentration; IPCC AR6 Table 7.15. This number includes indirect effects.

    # for conversion from ppb to kg-CH4
    M_ch4 = 16.04  # g/mol
    M_air = 28.97  # g/mol, dry air
    m_atmosphere = 5.135e18  # kg [Trenberth and Smith, 2005]

    radiative_efficiency_kg = (
        radiative_efficiency_ppb * M_air / M_ch4 * 1e9 / m_atmosphere
    )  # W/m2/kg-CH4

    tau = 11.8  # Lifetime (years)

    date_beginning: np.datetime64 = series["date"].to_numpy()
    date_characterized: np.ndarray = date_beginning + np.arange(
        start=0, stop=period, dtype="timedelta64[Y]"
    ).astype("timedelta64[s]")

    decay_multipliers: list = np.array(
        [
            radiative_efficiency_kg * tau * (1 - np.exp(-year / tau))
            for year in range(period)
        ]
    )

    forcing = pd.Series(data=series.amount * decay_multipliers, dtype="float64")

    if not cumulative:
        forcing = forcing.diff(periods=1).fillna(0)

    return pd.DataFrame(
        {
            "date": pd.Series(data=date_characterized, dtype="datetime64[s]"),
            "amount": forcing,
            "flow": series.flow,
            "activity": series.activity,
        }
    )


def characterize_n2o(
    series,
    period: int = 100,
    cumulative=False,
) -> pd.DataFrame:
    """
    Calculate the cumulative or marginal radiative forcing (CRF) from N2O for each year in a given period.
    
    Based on characterize_methane from bw_temporalis, but updated numerical values from IPCC AR6 Ch7 & SM.
   
    If `cumulative` is True, the cumulative CRF is calculated. If `cumulative` is False, the marginal CRF is calculated.
    Takes a single row of the TimeSeries Pandas DataFrame (corresponding to a set of (`date`/`amount`/`flow`/`activity`).
    For earch year in the given period, the CRF is calculated.
    Units are watts/square meter/kilogram of N2O.

    Parameters
    ----------
    series : array-like
        A single row of the TimeSeries dataframe.
    period : int, optional
        Time period for calculation (number of years), by default 100
    cumulative : bool, optional
        Should the RF amounts be summed over time?

    Returns
    -------
    A TimeSeries dataframe with the following columns:
    - date: datetime64[s]
    - amount: float
    - flow: str
    - activity: str

    See also
    --------
    Joos2013: Relevant scientific publication on CRF: https://doi.org/10.5194/acp-13-2793-2013
    Schivley2015: Relevant scientific publication on the numerical calculation of CRF: https://doi.org/10.1021/acs.est.5b01118
    Forster2023: Updated numerical values from IPCC AR6 Chapter 7 (Table 7.15): https://doi.org/10.1017/9781009157896.009
    """

    # functional variables and units (from publications listed in docstring)
    radiative_efficiency_ppb = 2.8e-3  # # W/m2/ppb; 2019 background cch4 concentration; IPCC AR6 Table 7.15. This number includes indirect effects.

    # for conversion from ppb to kg-CH4
    M_n2o = 44.01  # g/mol
    M_air = 28.97  # g/mol, dry air
    m_atmosphere = 5.135e18  # kg [Trenberth and Smith, 2005]

    radiative_efficiency_kg = (
        radiative_efficiency_ppb * M_air / M_n2o * 1e9 / m_atmosphere
    )  # W/m2/kg-N2O

    tau = 109  # Lifetime (years)

    date_beginning: np.datetime64 = series["date"].to_numpy()
    date_characterized: np.ndarray = date_beginning + np.arange(
        start=0, stop=period, dtype="timedelta64[Y]"
    ).astype("timedelta64[s]")

    decay_multipliers: list = np.array(
        [
            radiative_efficiency_kg * tau * (1 - np.exp(-year / tau))
            for year in range(period)
        ]
    )

    forcing = pd.Series(data=series.amount * decay_multipliers, dtype="float64")
    if not cumulative:
        forcing = forcing.diff(periods=1).fillna(0)

    return pd.DataFrame(
        {
            "date": pd.Series(data=date_characterized, dtype="datetime64[s]"),
            "amount": forcing,
            "flow": series.flow,
            "activity": series.activity,
        }
    )


def create_generic_characterization_function(decay_series) -> pd.DataFrame:
    """
    Creates a characterization function for a GHG based on a decay series, by calling the nested method `characterize_generic()`.
    
    Parameters
    ----------
    decay_series : np.ndarray
        A decay series for a specific GHG. This is retrieved from `../data/decay_multipliers.pkl`

    Returns
    -------
    A function called `characterize_generic`, which in turn returns a TimeSeries dataframe that contains the forcing of the emission of the row over the given period based on the decay series of that biosphere flow.

    """

    def characterize_generic(
        series,
        period: int = 100,
        cumulative=False,
    ) -> pd.DataFrame:
        """
        Uses lookup generated in /dev/calculate_metrics.ipynb
        Data originates from https://doi.org/10.1029/2019RG000691

        Parameters
        ----------
        series : array-like
            A single row of the dynamic inventory dataframe.
        period : int, optional
            Time period for calculation (number of years), by default 100
        cumulative : bool,
            cumulative impact

        Returns
        -------
        A TimeSeries dataframe that contains the forcing of the point emission from the row for each year in the given period.
          date: datetime64[s]
          amount: float (forcing at this timestep)
          flow: str
          activity: str
          
        See also
        --------
        Joos2013: Relevant scientific publication on CRF: https://doi.org/10.5194/acp-13-2793-2013
        Schivley2015: Relevant scientific publication on the numerical calculation of CRF: https://doi.org/10.1021/acs.est.5b01118
        Forster2023: Updated numerical values from IPCC AR6 Chapter 7 (Table 7.15): https://doi.org/10.1017/9781009157896.009

        """

        date_beginning: np.datetime64 = series["date"].to_numpy()

        dates_characterized: np.ndarray = date_beginning + np.arange(
            start=0, stop=period, dtype="timedelta64[Y]"
        ).astype("timedelta64[s]")

        decay_multipliers = decay_series[:period]

        forcing = pd.Series(data=series.amount * decay_multipliers, dtype="float64")

        if not cumulative:
            forcing = forcing.diff(periods=1).fillna(0)

        return pd.DataFrame(
            {
                "date": pd.Series(data=dates_characterized, dtype="datetime64[s]"),
                "amount": forcing,
                "flow": series.flow,
                "activity": series.activity,
            }
        )

    return characterize_generic



from dataclasses import dataclass
from heapq import heappop, heappush
from typing import Callable, List
from numbers import Number
import numpy as np
from bw_temporalis import TemporalisLCA, TemporalDistribution

datetime_type = np.dtype("datetime64[s]")
timedelta_type = np.dtype("timedelta64[s]")


@dataclass
class Edge:
    """
    Class for storing a temporal edge with source and target.

    Leaf edges link to a source process which is a leaf in
    our graph traversal (either through cutoff or a filter
    function).

    """

    distribution: TemporalDistribution
    leaf: bool
    consumer: int
    producer: int
    td_producer: TemporalDistribution
    td_consumer: TemporalDistribution
    abs_td_producer: TemporalDistribution = None
    abs_td_consumer: TemporalDistribution = None


class EdgeExtractor(TemporalisLCA):
    """
    Child class of TemporalisLCA that traverses the supply chain just as the parent class but can create a timeline of edges, in addition timeline of flows or nodes.
    The edge timeline is then used to match the timestamp of edges to that of background databases and to replace these edges with edges from these background databases using Brightway Datapackages.
    """

    def __init__(self, *args, edge_filter_function: Callable = None, **kwargs) -> None:
        """
        Initialize the EdgeExtractor class.

        Parameters
        ----------
        *args : Variable length argument list
        edge_filter_function : Callable, optional
            A callable that filters edges. If not provided, a function that always returns False is used.
        **kwargs : Arbitrary keyword arguments

        Returns
        -------
        None
        
        """
        super().__init__(*args, **kwargs) #use __init__ of TemporalisLCA
        if edge_filter_function:
            self.edge_ff = edge_filter_function
        else:
            self.edge_ff = lambda x: False

    def build_edge_timeline(self) -> list:
        """
        Creates a timeline of the edges from the graph traversal. Starting from the edges of the functional unit node, it goes through each node using a heap, selecting the node with the highest impact first.
        It, then, propagates the TemporalDistributions of the edges from node to node through time using convolution-operators during multiplication.
        It stops in case the current edge is known to have no temporal distribution (=leaf) (e.g. part of background database).

        Parameters
        ----------
        None

        Returns
        -------
        list
            A list of Edge instances with timestamps and amounts, and ids of its producing and consuming node.

        """

        heap = []
        timeline = []

        for edge in self.edge_mapping[
            self.unique_id
        ]:  # starting at the edges of the functional unit
            node = self.nodes[edge.producer_unique_id]
            heappush(
                heap,
                (
                    1 / node.cumulative_score,
                    self.t0 * edge.amount,
                    self.t0,
                    self.t0,
                    node,
                ),
            )
            timeline.append(
                Edge(
                    distribution=self.t0 * edge.amount,
                    leaf=False,
                    consumer=self.unique_id,
                    producer=node.activity_datapackage_id,
                    td_producer=edge.amount,
                    td_consumer=self.t0,
                    abs_td_producer=self.t0,
                )
            )

        while heap:
            _, td, td_parent, abs_td, node = heappop(heap)

            for edge in self.edge_mapping[node.unique_id]:
                row_id = self.nodes[edge.producer_unique_id].activity_datapackage_id
                col_id = node.activity_datapackage_id
                exchange = self.get_technosphere_exchange(
                    input_id=row_id,
                    output_id=col_id,
                )
                td_producer = (  # td_producer is the TemporalDistribution of the edge
                    self._exchange_value(
                        exchange=exchange,
                        row_id=row_id,
                        col_id=col_id,
                        matrix_label="technosphere_matrix",
                    )
                    / node.reference_product_production_amount
                )
                producer = self.nodes[edge.producer_unique_id]
                leaf = self.edge_ff(row_id)

                # If an edge does not have a TD, give it a td with timedelta=0 and the amount= 'edge value'
                if isinstance(td_producer, Number):
                    td_producer = TemporalDistribution(
                        date=np.array([0], dtype="timedelta64[Y]"),  
                        amount=np.array([td_producer]),
                    )

                distribution = (
                    td * td_producer
                ).simplify()  # convolution-multiplication of TemporalDistribution of consuming node (td) and consumed edge (edge) gives TD of producing node
                
                timeline.append(
                    Edge(
                        distribution=distribution,
                        leaf=leaf,
                        consumer=node.activity_datapackage_id,
                        producer=producer.activity_datapackage_id,
                        td_producer=td_producer,
                        td_consumer=td_parent,
                        abs_td_producer=self.join_datetime_and_timedelta_distributions(
                            td_producer, abs_td
                        ),
                        abs_td_consumer=abs_td,
                    )
                )
                if not leaf:
                    heappush(
                        heap,
                        (
                            1 / node.cumulative_score,
                            distribution,
                            td_producer,
                            self.join_datetime_and_timedelta_distributions(
                                td_producer, abs_td
                            ),
                            producer,
                        ),
                    )
        return timeline

    def join_datetime_and_timedelta_distributions(
        self, td_producer: TemporalDistribution, td_consumer: TemporalDistribution
    ) -> TemporalDistribution:
        """
        Joins a datetime TemporalDistribution (td_producer) with a timedelta TemporalDistribution (td_consumer) to create a new TemporalDistribution.

        If the producer does not have a TemporalDistribution, the consumer's TemporalDistribution is returned to continue the timeline.
        If both the producer and consumer have TemporalDistributions, they are joined together.

        Parameters
        ----------
        td_producer : TemporalDistribution
            TemporalDistribution of the producer. Expected to be a datetime TemporalDistribution.
        td_consumer : TemporalDistribution
            TemporalDistribution of the consumer. Expected to be a timedelta TemporalDistribution.

        Returns
        -------
        TemporalDistribution
            A new TemporalDistribution that is the result of joining the producer and consumer TemporalDistributions.

        Raises
        ------
        ValueError
            If the dtype of `td_consumer.date` is not `datetime64[s]` or the dtype of `td_producer.date` is not `timedelta64[s]`.

        """
        # if an edge does not have a TD, then return the consumer_td so that the timeline continues
        if isinstance(td_consumer, TemporalDistribution) and isinstance(
            td_producer, Number
        ):
            return td_consumer
        
        # Else, if both consumer and producer have a td (absolute and relative, respectively) join to TDs
        if isinstance(td_producer, TemporalDistribution) and isinstance(
            td_consumer, TemporalDistribution
        ):
            if not (td_consumer.date.dtype == datetime_type):
                raise ValueError(
                    f"`td_consumer.date` must have dtype `datetime64[s]`, but got `{td_consumer.date.dtype}`"
                )
            if not (td_producer.date.dtype == timedelta_type):
                raise ValueError(
                    f"`td_producer.date` must have dtype `timedelta64[s]`, but got `{td_producer.date.dtype}`"
                )
            date = (
                td_consumer.date.reshape((-1, 1)) + td_producer.date.reshape((1, -1))
            ).ravel()
            amount = np.array(len(td_consumer) * [td_producer.amount]).ravel()
            return TemporalDistribution(date, amount)
        else:
            raise ValueError(
                "Can't join TemporalDistribution and something else: Trying with {} and {}".format(
                    type(td_consumer.date), type(td_producer.date)
                )
            )



import bw2data as bd
import warnings
import pandas as pd
import numpy as np
from typing import Union, Tuple, Optional, Callable
from datetime import datetime, timedelta
from typing import KeysView

from bw_temporalis import TemporalDistribution
from bw2calc import LCA
from .edge_extractor import EdgeExtractor, Edge
from .utils import (
    extract_date_as_integer,
    extract_date_as_string,
    convert_date_string_to_datetime,
)


class TimelineBuilder:
    """
    This class is responsible for building a timeline of processes based on the temporal relationship of the priority-first graph traversal.

    First, the `EdgeExtractor` is called and it extracts a timeline of exchanges (edge_timeline) with tempoal information.
    Identical edges within temporal grouping (e.g. year, month, day, hour) are then grouped together and the amount of exchanges is summed up.
    """

    def __init__(
        self,
        slca: LCA,
        edge_filter_function: Callable,
        database_date_dict: dict,
        database_date_dict_static_only: dict,
        time_mapping_dict: dict,
        node_id_collection_dict: dict,
        temporal_grouping: str = "year",
        interpolation_type: str = "linear",
        cutoff: float = 1e-9,
        max_calc: float = 1e4,
        *args,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        slca: LCA
            A static LCA object.
        edge_filter_function: Callable
            A callable that filters edges. If not provided, a function that always returns False is used.
        database_date_dict: dict
            A dictionary mapping databases to dates.
        time_mapping_dict: dict
          A dictionary to map processes to specific times.
        temporal_grouping: str, optional
            The temporal grouping to be used. Default is "year".
        interpolation_type: str, optional
            The type of interpolation to be used to select the background databases. Default is "linear".
        cutoff: 
            The cutoff value for the graph traversal. Default is 1e-9.
        max_calc: 
            The maximum number of calculations to be performed by the graph traversal. Default is 1e4.
        args:   Variable length argument list
            Keyword arguments passed to the EdgeExtractor which inherits from TemporalisLCA. Here, things like the further settings for graph traversal can be set. For details, see bw_temporalis.TemporalisLCA.
        kwargs: Arbitrary keyword arguments
            Keyword arguments passed to the EdgeExtractor which inherits from TemporalisLCA.
        """
        self.slca = slca
        self.edge_filter_function = edge_filter_function
        self.database_date_dict = database_date_dict
        self.database_date_dict_static_only = database_date_dict_static_only
        self.time_mapping_dict = time_mapping_dict
        self.node_id_collection_dict = node_id_collection_dict
        self.temporal_grouping = temporal_grouping
        self.interpolation_type = interpolation_type
        self.cutoff = cutoff
        self.max_calc = max_calc

        # Finding indices of activities from background databases that are known to be static, i.e. have no temporal distributions connecting to them. These will be be skipped in the graph traversal.
        static_activity_db_indices = [
            node_id
            for node_id in self.node_id_collection_dict[
                "demand_dependent_background_node_ids"
            ]
            if node_id
            not in self.node_id_collection_dict[
                "first_level_background_node_ids_static"
            ]
        ]

        # The graph traversal needs the matrix indices, not the database / node indices.
        static_activity_matrix_indices = [
            self.slca.dicts.product[node_id] for node_id in static_activity_db_indices
        ]

        self.edge_extractor = EdgeExtractor(
            slca,
            *args,
            edge_filter_function=edge_filter_function,
            cutoff=self.cutoff,
            max_calc=self.max_calc,
            static_activity_indices=set(static_activity_matrix_indices),
            **kwargs,
        )
        self.edge_timeline = self.edge_extractor.build_edge_timeline()

    def check_database_names(self) -> None:
        """
        Check that the strings of the databases exist in the databases of the brightway project.

        """
        for db in self.database_date_dict_static_only.keys():
            assert (
                db in bd.databases
            ), f"{db} is not in your brightway project databases."
        return

    # TODO: rethink structure of build_timeline(): is it good to have all these nested functions?
    def build_timeline(self) -> pd.DataFrame:
        """
        Create a dataframe with grouped, time-explicit edges and for each grouped edge interpolate to the database with the closest time of representativeness.

        Edges from same producer to same consumer that occur at different times within the same time window (temporal_grouping) are grouped together. Possible temporal groupings are "year", "month", "day" and "hour".

        The column "interpolation weights" assigns the ratio [0-1] of the edge's amount to be taken from the database with the closest time of representativeness.
        Available interpolation types are:
            - "linear": linear interpolation between the two closest databases, based on temporal distance.
            - "closest": closest database is assigned 1

        Parameters
        ----------
        None 
            (all are already passed during instantiation)
        
        Returns
        -------
        pd.DataFrame
            A timeline with grouped, time-explicit edges and interpolation weights to background databases.
        """

        def extract_edge_data(edge: Edge) -> dict:
            """
            Stores the attributes of an Edge instance in a dictionary.

            :param edge: Edge instance
            :return: Dictionary with attributes of the edge
            """
            try:
                consumer_date = edge.abs_td_consumer.date
                consumer_date = np.array(
                    [consumer_date for i in range(len(edge.td_producer))]
                ).T.flatten()
            except AttributeError:
                consumer_date = None

            return {
                "producer": edge.producer,
                "consumer": edge.consumer,
                "leaf": edge.leaf,
                "consumer_date": consumer_date,
                "producer_date": edge.abs_td_producer.date,
                "amount": edge.abs_td_producer.amount,
            }

        def get_consumer_name(id: int) -> str:
            """
            Returns the name of consumer node.
            If consuming node is the functional unit, returns -1.

            :param id: Id of node
            :return: string of node's name or -1
            """
            try:
                return bd.get_node(id=id)["name"]
            except:
                return "-1"  # functional unit

        def add_column_interpolation_weights_to_timeline(
            tl_df: pd.DataFrame,
            interpolation_type: str = "linear",
        ) -> pd.DataFrame:
            """
            Add a column to a timeline with the weights for an interpolation between the two nearest dates, from the list of dates of the available databases.

            :param tl_df: Timeline as a dataframe.
            :param interpolation_type: Type of interpolation between the nearest lower and higher dates. Available options: "linear"and "nearest".

            :return: Timeline as a dataframe with a column 'interpolation_weights' added, this column looks like {database_name: weight, database_name: weight}.

            """
            if not self.database_date_dict_static_only:
                tl_df["interpolation_weights"] = None
                warnings.warn(
                    "No time-explicit databases are provided. Mapping to time-explicit databases is not possible.",
                    category=Warning,
                )
                return tl_df

            dates_list = [
                date
                for date in self.database_date_dict_static_only.values()
                if type(date) == datetime
            ]
            if "date_producer" not in list(tl_df.columns):
                raise ValueError("The timeline does not contain dates.")

            # create reversed dict {date: database} with only static "background" db's
            self.reversed_database_date_dict = {
                v: k
                for k, v in self.database_date_dict_static_only.items()
                if type(v) == datetime
            }

            if self.interpolation_type == "nearest":
                tl_df["interpolation_weights"] = tl_df["date_producer"].apply(
                    lambda x: find_closest_date(x, dates_list)
                )

            if self.interpolation_type == "linear":
                tl_df["interpolation_weights"] = tl_df["date_producer"].apply(
                    lambda x: get_weights_for_interpolation_between_nearest_years(
                        x, dates_list, self.interpolation_type
                    )
                )

            else:
                raise ValueError(
                    f"Sorry, but {self.interpolation_type} interpolation is not available yet."
                )

            tl_df["interpolation_weights"] = tl_df.apply(
                add_interpolation_weights_at_intersection_to_background, axis=1
            )  # add the weights to the timeline for processes at intersection

            return tl_df

        def add_interpolation_weights_at_intersection_to_background(row) -> Union[dict, None]:
            """
            returns the interpolation weights to background databases only for those exchanges, where the producing process
            actually comes from a background database (temporal markets).

            Only these processes are receiving inputs from the background databases.
            All other process in the timeline are not directly linked to the background, so the interpolation weight info is not needed and set to None

            Parameters
            ----------
            row : pd.Series
                Row of the timeline dataframe
            Returns
            -------
            dict
                Dictionary with the name of databases and interpolation weights.
            """

            if (
                row["producer"]
                in self.node_id_collection_dict[
                    "first_level_background_node_ids_static"
                ]
            ):
                return {
                    self.reversed_database_date_dict[x]: v
                    for x, v in row["interpolation_weights"].items()
                }
            else:
                return None

        def find_closest_date(target: datetime, dates: KeysView[datetime]) -> dict:
            """
            Find the closest date to the target in the dates list.

            :param target: Target datetime.datetime object.
            :param dates: List of datetime.datetime objects.
            :return: Dictionary with the key as the closest datetime.datetime object from the list and a value of 1.

            """

            # If the list is empty, return None
            if not dates:
                return None

            # Sort the dates
            dates = sorted(dates)
            # Use min function with a key based on the absolute difference between the target and each date
            closest = min(dates, key=lambda date: abs(target - date))

            return {closest: 1}

        def get_weights_for_interpolation_between_nearest_years(
            reference_date: datetime,
            dates_list: KeysView[datetime],
            interpolation_type: str = "linear",
        ) -> dict:
            """
            Find the nearest dates (before and after) a given date in a list of dates and calculate the interpolation weights.

            :param reference_date: Target date.
            :param dates_list: KeysView[datetime], which is a list of the temporal representativeness of the available databases.
            :param interpolation_type: Type of interpolation between the nearest lower and higher dates. For now,
            only "linear" is available.

            :return: Dictionary with temporal coverage of the available databases to use as keys and the weights for interpolation as values.

            """
            dates_list = sorted(dates_list)

            diff_dates_list = [reference_date - x for x in dates_list]

            if timedelta(0) in diff_dates_list:  # date of process == date of database
                exact_match = dates_list[diff_dates_list.index(timedelta(0))]
                return {exact_match: 1}

            closest_lower = None
            closest_higher = None

            # select the closest lower and higher dates of the database in regards to the date of process
            for date in dates_list:
                if date < reference_date:
                    if (
                        closest_lower is None
                        or reference_date - date < reference_date - closest_lower
                    ):
                        closest_lower = date
                elif date > reference_date:
                    if (
                        closest_higher is None
                        or date - reference_date < closest_higher - reference_date
                    ):
                        closest_higher = date

            if closest_lower is None:
                warnings.warn(
                    f"Reference date {reference_date} is lower than all provided dates. Data will be taken from the closest higher year.",
                    category=Warning,
                )
                return {closest_higher: 1}

            if closest_higher is None:
                warnings.warn(
                    f"Reference date {reference_date} is higher than all provided dates. Data will be taken from the closest lower year.",
                    category=Warning,
                )
                return {closest_lower: 1}

            if self.interpolation_type == "linear":
                weight = int((reference_date - closest_lower).total_seconds()) / int(
                    (closest_higher - closest_lower).total_seconds()
                )
            else:
                raise ValueError(
                    f"Sorry, but {interpolation_type} interpolation is not available yet."
                )
            return {closest_lower: 1 - weight, closest_higher: weight}

        # check if database names match with databases in BW project
        self.check_database_names()

        # Check if temporal_grouping is a valid value
        valid_temporal_groupings = ["year", "month", "day", "hour"]
        if self.temporal_grouping not in valid_temporal_groupings:
            raise ValueError(
                f"Invalid value for 'temporal_grouping'. Allowed values are {valid_temporal_groupings}."
            )

        # Extract edge data into a list of dictionaries
        edges_data = [extract_edge_data(edge) for edge in self.edge_timeline]

        # Convert list of dictionaries to dataframe
        edges_df = pd.DataFrame(edges_data)

        # Explode datetime and amount columns: each row with multiple dates and amounts is exploded into multiple rows with one date and one amount
        edges_df = edges_df.explode(["consumer_date", "producer_date", "amount"])
        edges_df.drop_duplicates(inplace=True)
        edges_df = edges_df[edges_df["amount"] != 0]

        # For the Functional Unit: set consumer date = producer date as it occurs at the same time
        edges_df.loc[edges_df["consumer"] == -1, "consumer_date"] = edges_df.loc[
            edges_df["consumer"] == -1, "producer_date"
        ]

        # extract grouping time of consumer and producer: processes occuring at different times withing in teh same time window of grouping get the same grouping time
        edges_df["consumer_grouping_time"] = edges_df["consumer_date"].apply(
            lambda x: extract_date_as_string(self.temporal_grouping, x)
        )
        edges_df["producer_grouping_time"] = edges_df["producer_date"].apply(
            lambda x: extract_date_as_string(self.temporal_grouping, x)
        )

        # group unique pair of consumer and producer with the same grouping times
        grouped_edges = (
            edges_df.groupby(
                [
                    "producer_grouping_time",
                    "consumer_grouping_time",
                    "producer",
                    "consumer",
                ]
            )
            .agg({"amount": "sum"})
            .reset_index()
        )

        # convert grouping times, which was only used as intermediate variable, back to datetime
        grouped_edges["date_producer"] = grouped_edges["producer_grouping_time"].apply(
            lambda x: convert_date_string_to_datetime(
                self.temporal_grouping, x
            )
        )
        grouped_edges["date_consumer"] = grouped_edges["consumer_grouping_time"].apply(
            lambda x: convert_date_string_to_datetime(
                self.temporal_grouping, x
            )
        )

        # add dates as integers as hashes to the dataframe
        grouped_edges["hash_producer"] = grouped_edges["date_producer"].apply(
            lambda x: extract_date_as_integer(x, time_res=self.temporal_grouping)
        )

        grouped_edges["hash_consumer"] = grouped_edges["date_consumer"].apply(
            lambda x: extract_date_as_integer(x, time_res=self.temporal_grouping)
        )

        # add new processes to time mapping dict
        for row in grouped_edges.itertuples():
            self.time_mapping_dict.add(
                (bd.get_node(id=row.producer).key, row.hash_producer)
            )

        # store the ids from the time_mapping_dict in dataframe
        grouped_edges["time_mapped_producer"] = grouped_edges.apply(
            lambda row: self.time_mapping_dict[
                (bd.get_node(id=row.producer).key, row.hash_producer)
            ],
            axis=1,
        )

        grouped_edges["time_mapped_consumer"] = grouped_edges.apply(
            lambda row: (
                self.time_mapping_dict[
                    (bd.get_node(id=row.consumer).key, row.hash_consumer)
                ]
                if row.consumer != -1
                else -1
            ),
            axis=1,
        )

        # Add interpolation weights to background databases to the dataframe
        grouped_edges = add_column_interpolation_weights_to_timeline(
            grouped_edges,
            interpolation_type=self.interpolation_type,
        )

        # Retrieve producer and consumer names
        grouped_edges["producer_name"] = grouped_edges.producer.apply(
            lambda x: bd.get_node(id=x)["name"]
        )
        grouped_edges["consumer_name"] = grouped_edges.consumer.apply(get_consumer_name)

        # Reorder columns
        grouped_edges = grouped_edges[
            [
                "hash_producer",
                "time_mapped_producer",
                "date_producer",
                "producer",
                "producer_name",
                "hash_consumer",
                "time_mapped_consumer",
                "date_consumer",
                "consumer",
                "consumer_name",
                "amount",
                "interpolation_weights",
            ]
        ]

        return grouped_edges



import bw2data as bd
import bw_processing as bwp
import uuid
import numpy as np
import pandas as pd
from typing import Optional
from datetime import datetime


class MatrixModifier:
    """
    This class is responsible for modifying the original LCA matrices to contain the time-explicit processes and to relink them to the time-explicit background databasess. 
    
    It does this by ccreating datapackages that contain new matrix entries all changes in the matrices, based on a timeline dataframe (created from TimelineBuilder.build_timeline()).

    """

    def __init__(
        self,
        timeline: pd.DataFrame,
        database_date_dict_static_only: dict,
        demand_timing: dict,
        name: Optional[str] = None,
    ) -> None:
        """
        Initializes the MatrixModifier object.

        Parameters
        ----------
        timeline : pd.DataFrame
            A DataFrame of the timeline of exchanges
        database_date_dict_static_only : dict
            A dictionary mapping the static background databases to dates.
        demand_timing : dict
            A dictionary mapping the demand to its timing.
        name : str, optional   
            An optional name for the MatrixModifier instance. Default is None.
        temporalized_process_ids : set
            A set to collect the ids the "exploded" processes, instantiated empty
        temporal_market_ids : set
            A set to collect the ids of the "temporal markets" , instantiated empty
        """

        self.timeline = timeline
        self.database_date_dict_static_only = database_date_dict_static_only
        self.demand_timing = demand_timing
        self.name = name
        self.temporalized_process_ids = set()
        self.temporal_market_ids = set()

    def create_technosphere_datapackage(self) -> bwp.Datapackage:
        """
        Creates patches to the technosphere matrix from a given timeline of grouped exchanges to add these temporal processes to the technopshere database.
        Patches are datapackages that add or overwrite datapoints in the LCA matrices before LCA calculations.

        The heavy lifting of this function happens in its inner function `add_row_to_datapackage()`:
        Here, each node with a temporal distribution is "exploded", which means each occurrence of this node (e.g. steel production on 2020-01-01
        and steel production 2015-01-01) becomes a separate, time-explicit new node, by adding the respective elements to the technosphere matrix.
        For processes at the interface with background databases, the timing of the exchanges determines which background database to link to in so called "Temporal Markets".

        Parameters
        ----------
        None

        Returns
        -------
        bwp.Datapackage
            A datapackage containing the patches for the technosphere matrix.
        """

        def add_row_to_datapackage(
            row: pd.core.frame,
            datapackage: bwp.Datapackage,
            database_date_dict_static_only: dict,
            demand_timing: dict,
            new_nodes: set,
        ) -> None:
            """
            This adds the required technosphere matrix modifications for each time-dependent exchange (edge) as datapackage elements to a given `bwp.Datapackage`.
            Modifications include:
            1) Exploded processes: new matrix elements between exploded consumer and exploded producer, representing the temporal edge between them.
            2) Temporal markets: new matrix entries between "temporal markets" and the producers in temporally matching background databases, with shares based on interpolation.
               Processes in the background databases are matched on name, reference product and location.
            3) Diagonal entries: ones on the diagonal for new nodes.

            This function also updates the sets of new nodes with the ids of any new nodes created during this process.

            Parameters
            ----------
            row : pd.core.frame
                A row of the timeline DataFrame representing an temporalized edge
            datapackage : bwp.Datapackage
                Append to this datapackage, if available. Otherwise create a new datapackage.
            database_date_dict_static_only : dict
                A dict of the available background databases: their names (key) and temporal representativeness (value).
            demand_timing : dict
                A dict of the demand ids and their timing. Can be created using `TimexLCA.create_demand_timing_dict()`.
            new_nodes : set
                Set to which new node ids are added.

            Returns
            -------
            None but updates the set new_nodes and adds a patch for this new edge to the bwp.Datapackage.
            """

            if row.consumer == -1:  # functional unit
                new_producer_id = row.time_mapped_producer
                new_nodes.add(new_producer_id)
                self.temporalized_process_ids.add(
                    new_producer_id
                )  # comes from foreground, so it is a temporalized process
                return

            new_consumer_id = row.time_mapped_consumer
            new_nodes.add(new_consumer_id)

            new_producer_id = row.time_mapped_producer
            new_nodes.add(new_producer_id)

            previous_producer_id = row.producer
            previous_producer_node = bd.get_node(
                id=previous_producer_id
            )  # in future versions, insead of getting node, just provide list of producer ids

            # Add entry between exploded consumer and exploded producer (not in background database)
            datapackage.add_persistent_vector(
                matrix="technosphere_matrix",
                name=uuid.uuid4().hex,
                data_array=np.array([row.amount], dtype=float),
                indices_array=np.array(
                    [(new_producer_id, new_consumer_id)],
                    dtype=bwp.INDICES_DTYPE,
                ),
                flip_array=np.array(
                    [True], dtype=bool
                ),  
            )

            # Check if previous producer comes from background database
            if (
                previous_producer_node["database"]
                in self.database_date_dict_static_only.keys()
            ):
                # Create new edges based on interpolation_weights from the row
                for database, db_share in row.interpolation_weights.items():
                    # Get the producer activity in the corresponding background database
                    try:
                        producer_id_in_background_db = bd.get_node(
                            **{
                                "database": database,
                                "name": previous_producer_node["name"],
                                "product": previous_producer_node["reference product"],
                                "location": previous_producer_node["location"],
                            }
                        ).id
                    except:
                        print(
                            f"Could not find producer in database {database} with name {previous_producer_node['name']}, product {previous_producer_node['reference product']}, location {previous_producer_node['location']}"
                        )
                        raise SystemExit

                    # Add entry between exploded producer and producer in background database ("Temporal Market")
                    datapackage.add_persistent_vector(
                        matrix="technosphere_matrix",
                        name=uuid.uuid4().hex,
                        data_array=np.array(
                            [db_share], dtype=float
                        ),  # temporal markets produce 1, so shares divide amount between dbs
                        indices_array=np.array(
                            [(producer_id_in_background_db, new_producer_id)],
                            dtype=bwp.INDICES_DTYPE,
                        ),
                        flip_array=np.array([True], dtype=bool),
                    )
                    self.temporal_market_ids.add(new_producer_id)
            else:
                self.temporalized_process_ids.add(
                    new_producer_id
                )  # comes from foreground, so it is a temporalized process

        datapackage = bwp.create_datapackage(
            sum_inter_duplicates=False
        )  # 'sum_inter_duplicates=False': If the same market is used by multiple foreground processes, the market gets created again, inputs should not be summed.

        new_nodes = set()

        for row in self.timeline.iloc[::-1].itertuples():
            add_row_to_datapackage(
                row,
                datapackage,
                self.database_date_dict_static_only,
                self.demand_timing,
                new_nodes,
            )

        # Adding ones on diagonal for new nodes
        datapackage.add_persistent_vector(
            matrix="technosphere_matrix",
            name=uuid.uuid4().hex,
            data_array=np.ones(len(new_nodes)),
            indices_array=np.array(
                [(i, i) for i in new_nodes], dtype=bwp.INDICES_DTYPE
            ),
        )

        return datapackage

    def create_biosphere_datapackage(self) -> bwp.Datapackage:
        """
        Creates list of patches formatted as datapackages for modifications to the biosphere matrix.
        It adds the biosphere flows to the exploded technosphere processes.

        This function iterates over each unique producer and for each biosphere exchange of the original activity,
        it creates a new biosphere exchange for the new node.

        Temporal markets have to biosphere exchanges, as they only divide the amount of the technosphere exchange between the different databases.

        Parameters
        ----------
        None

        Returns
        -------
        bwp.Datapackage
            An updated datapackage containing the patches for the biosphere matrix.
        """
        unique_producers = (
            self.timeline.groupby(["producer", "time_mapped_producer"])
            .count()
            .index.values
        )  # array of unique (producer, timestamp) tuples

        datapackage_bio = bwp.create_datapackage(sum_inter_duplicates=False)

        for producer in unique_producers:
            if (
                bd.get_activity(producer[0])["database"]
                not in self.database_date_dict_static_only.keys()  # skip temporal markets
            ):
                producer_id = producer[1]
                # the producer_id is a combination of the activity_id and the timestamp
                producer_node = bd.get_node(id=producer[0])
                indices = (
                    []
                )  # list of (biosphere, technosphere) indices for the biosphere flow exchanges
                amounts = []  # list of amounts corresponding to the bioflows
                for exc in producer_node.biosphere():
                    indices.append(
                        (exc.input.id, producer_id)
                    )  # directly build a list of tuples to pass into the datapackage, the producer_id is used to for the column of that activity
                    amounts.append(exc.amount)

                datapackage_bio.add_persistent_vector(
                    matrix="biosphere_matrix",
                    name=uuid.uuid4().hex,
                    data_array=np.array(amounts, dtype=float),
                    indices_array=np.array(
                        indices,
                        dtype=bwp.INDICES_DTYPE,
                    ),
                    flip_array=np.array([False], dtype=bool),
                )
        return datapackage_bio

    def create_datapackage(self) -> None:
        """
        Creates a list of datapackages for the technosphere and biosphere matrices, by calling the respective functions.

        Parameters
        ----------
        None

        Returns
        -------
        list
            A list of the technosphere and biosphere datapackages.
        """
        technosphere_datapackage = self.create_technosphere_datapackage()
        biosphere_datapackage = self.create_biosphere_datapackage()
        
        return [technosphere_datapackage, biosphere_datapackage]



"""
unit tests for temporal groupings:
- Years
- months
- missing test: days, hours: such long integers, e.g. 20240123 for 23.01.2024 or 2023012314 for 14:00 on 23.01.2024, 
cannot be added to the new ids of the process copies because this makes them too large to be converted to C long
See issue on Github.

currently only with temporal distributions on 1 level, because of issue of finding the correct column id (issue #6)
"""

import math
import unittest
from bw_temporalis import easy_timedelta_distribution, TemporalDistribution
from edge_extractor import EdgeExtracter
from medusa_tools import *
import bw2data as bd
import bw2calc as bc
import numpy as np


class TestTemporalGrouping(unittest.TestCase):
    
    def test_temporal_grouping_months(self):
        # create a simple foreground database with 2 processes A and B, where B consumes C from background database
        # C is produced in background database in 2022 and 2024, with different CO2 emissions
        # A consumes B with a temporal distribution of 1/3, 1/3, 1/3
        # The expected result is a medusa LCA score of 12.5 for A
        # The test checks if the monthly aggregation and LCA score is calculated correctly
        
        def prepare_medusa_dbs():
            bd.projects.set_current("__test_medusa__")

            bd.Database('medusa-bio').write({
                        ('medusa-bio', "CO2"): {
                            "type": "biosphere",
                            "name": "carbon dioxide",
                            "temporalis code": "co2",
                        },
                    },
            )
            
            bd.Database("background_2024").write(
                {
                    ("background_2024", "C"): {
                        'name': 'C',
                        'location': 'somewhere',
                        'reference product': 'C',
                        "exchanges": [
                            {
                                'amount': 1,
                                'type': 'production',
                                'input': ("background_2024", 'C'),
                            },
                            
                            {
                                "amount": 10,
                                "input": ("medusa-bio", "CO2"),
                                "type": "biosphere",
                            },
                        ],       
                    },
                },
                
            )
            
            bd.Database("background_2022").write(
                {
                    ("background_2022", "C"): {
                        'name': 'C',
                        'location': 'somewhere',
                        'reference product': 'C',
                        "exchanges": [
                            {
                                'amount': 1,
                                'type': 'production',
                                'input': ("background_2022", 'C'),
                            },                   
                            {
                                "amount": 15,
                                "input": ("medusa-bio", "CO2"),
                                "type": "biosphere",
                            },
                        ],   
                    },
                },
                
            )
            

            
            bd.Database("foreground").write(
                {
                    ("foreground", "A"): {
                        'name': 'A',
                        'location': 'somewhere',
                        'reference product': 'A',
                        "exchanges": [
                            {
                                'amount': 1,
                                'type': 'production',
                                'input': ('foreground', 'A'),
                            },                 
                            {
                                "amount": 1,
                                "input": ("foreground", "B"),
                                'temporal_distribution': TemporalDistribution(
                                        np.array([-24, -12, 0], dtype='timedelta64[M]'),
                                        np.array([1/3, 1/3, 1/3])),  
                                "type": "technosphere",
                            },
                        ],
                    },
                    
                    ("foreground", "B"): {
                        'name': 'B',
                        'location': 'somewhere',
                        'reference product': 'B',
                        "exchanges": [
                            {
                                'amount': 1,
                                'type': 'production',
                                'input': ('foreground', 'B'),
                            },
                            {
                                "amount": 1,
                                "input": ("background_2024", "C"),
                                "type": "technosphere",
                            },
                        ],
                    },
                }
            )
        
            bd.Method(("GWP", "example")).write([
            (("medusa-bio", "CO2"), 1),
            ])    
            return 
        
        prepare_medusa_dbs()
        bd.projects.set_current("__test_medusa__") # monthly grouping
        bd.databases
        
        demand = {('foreground', 'A'): 1}
        method = ("GWP", "example")
        
        # calculate static LCA'
        slca = bc.LCA(demand, method)
        slca.lci()
        slca.lcia()     
        
        # calculate Medusa LCA
        SKIPPABLE = [node.id for node in bd.Database('background_2022')] + [
                    node.id for node in bd.Database('background_2024')
        ]

        def filter_function(database_id: int) -> bool:
            return database_id in SKIPPABLE
        
        eelca = EdgeExtracter(slca, edge_filter_function=filter_function)
        
        timeline = eelca.build_edge_timeline()
        
        database_date_dict = {
                datetime.strptime("2022-01", "%Y-%m"): 'background_2022',
                datetime.strptime("2024-01", "%Y-%m"): 'background_2024',        
            }

        timeline_df = create_grouped_edge_dataframe(timeline, database_date_dict, temporal_grouping= 'month', interpolation_type="linear") # monthly grouping

        demand_timing_dict = create_demand_timing_dict(timeline_df, demand)

        dp = create_datapackage_from_edge_timeline(timeline_df, database_date_dict, demand_timing_dict)
        
        fu, data_objs, remapping = prepare_medusa_lca_inputs(demand=demand, demand_timing_dict=demand_timing_dict, method=method) 
        
        lca = bc.LCA(fu, data_objs = data_objs + [dp], remapping_dicts=remapping)
        lca.lci()
        lca.lcia()
        
        #calculate expected score:
        expected_score = 1/3*15 +1/6*15 + 1/6*10 + 1/3*10
        
        print('\nTest Temporal Grouping Months:')
        print(timeline_df)
        print('Static LCA Score:', slca.score)
        print('MEDUSA LCA Score:', lca.score)
        print('Expected MEDUSA LCA Score:', expected_score)
        
        
        # Check if the results are equal using math.isclose
        self.assertTrue(math.isclose(lca.score, expected_score, rel_tol=1e-9))
        
        
    def test_temporal_grouping_years(self):
        # create a simple foreground database with 2 processes A and B, where B consumes C from background database
        # C is produced in background database in 2022 and 2024, with different CO2 emissions
        # A consumes B with a temporal distribution of 1/3, 1/3, 1/3
        # The expected result is a medusa LCA score of 12.5 for A
        # The test checks if the monthly aggregation and LCA score is calculated correctly
        
        def prepare_medusa_dbs():
            bd.projects.set_current("__test_medusa__")

            bd.Database('medusa-bio').write({
                        ('medusa-bio', "CO2"): {
                            "type": "biosphere",
                            "name": "carbon dioxide",
                            "temporalis code": "co2",
                        },
                    },
            )
            
            bd.Database("background_2024").write(
                {
                    ("background_2024", "C"): {
                        'name': 'C',
                        'location': 'somewhere',
                        'reference product': 'C',
                        "exchanges": [
                            {
                                'amount': 1,
                                'type': 'production',
                                'input': ("background_2024", 'C'),
                            },
                            
                            {
                                "amount": 10,
                                "input": ("medusa-bio", "CO2"),
                                "type": "biosphere",
                            },
                        ],       
                    },
                },
                
            )
            
            bd.Database("background_2022").write(
                {
                    ("background_2022", "C"): {
                        'name': 'C',
                        'location': 'somewhere',
                        'reference product': 'C',
                        "exchanges": [
                            {
                                'amount': 1,
                                'type': 'production',
                                'input': ("background_2022", 'C'),
                            },                   
                            {
                                "amount": 15,
                                "input": ("medusa-bio", "CO2"),
                                "type": "biosphere",
                            },
                        ],   
                    },
                },
                
            )
            
            bd.Database("foreground").write(
                {
                    ("foreground", "A"): {
                        'name': 'A',
                        'location': 'somewhere',
                        'reference product': 'A',
                        "exchanges": [
                            {
                                'amount': 1,
                                'type': 'production',
                                'input': ('foreground', 'A'),
                            },                 
                            {
                                "amount": 1,
                                "input": ("foreground", "B"),
                                'temporal_distribution': TemporalDistribution(
                                        np.array([-36, -24, -12, 0], dtype='timedelta64[M]'),
                                        np.array([1/4, 1/4, 1/4, 1/4])),  
                                "type": "technosphere",
                            },
                        ],
                    },
                    
                    ("foreground", "B"): {
                        'name': 'B',
                        'location': 'somewhere',
                        'reference product': 'B',
                        "exchanges": [
                            {
                                'amount': 1,
                                'type': 'production',
                                'input': ('foreground', 'B'),
                            },
                            {
                                "amount": 1,
                                "input": ("background_2024", "C"),
                                "type": "technosphere",
                            },
                        ],
                    },
                }
            )
        
            bd.Method(("GWP", "example")).write([
            (("medusa-bio", "CO2"), 1),
            ])    
            return 
        
        prepare_medusa_dbs()
        bd.projects.set_current("__test_medusa__") # monthly grouping
        bd.databases
        
        demand = {('foreground', 'A'): 1}
        method = ("GWP", "example")
        
        # calculate static LCA'
        slca = bc.LCA(demand, method)
        slca.lci()
        slca.lcia()       
        
        # calculate Medusa LCA
        SKIPPABLE = [node.id for node in bd.Database('background_2022')] + [
                    node.id for node in bd.Database('background_2024')
        ]

        def filter_function(database_id: int) -> bool:
            return database_id in SKIPPABLE
        
        eelca = EdgeExtracter(slca, edge_filter_function=filter_function)
        
        timeline = eelca.build_edge_timeline()
        
        database_date_dict = {
                datetime.strptime("2022-01", "%Y-%m"): 'background_2022',
                datetime.strptime("2024-01", "%Y-%m"): 'background_2024',        
            }

        timeline_df = create_grouped_edge_dataframe(timeline, database_date_dict, temporal_grouping= 'year', interpolation_type="linear") # yearly grouping

        demand_timing_dict = create_demand_timing_dict(timeline_df, demand)

        dp = create_datapackage_from_edge_timeline(timeline_df, database_date_dict, demand_timing_dict)
        
        fu, data_objs, remapping = prepare_medusa_lca_inputs(demand=demand, demand_timing_dict=demand_timing_dict, method=method) 
        
        lca = bc.LCA(fu, data_objs = data_objs + [dp], remapping_dicts=remapping)
        lca.lci()
        lca.lcia()
        
        #calculate expected score:
        expected_score = 1/4*15 + 1/4*15 +1/8*15 + 1/8*10 + 1/4*10
        
        print('\nTest Temporal Grouping Years:')     
        print(timeline_df)
        print('Static LCA Score:', slca.score)
        print('MEDUSA LCA Score:', lca.score)
        print('Expected MEDUSA LCA Score:', expected_score)
        
        
        # Check if the results are equal using math.isclose
        self.assertTrue(math.isclose(lca.score, expected_score, rel_tol=1e-9))

if __name__ == '__main__':
    unittest.main()

     

    

